from nacl.signing import VerifyKey, exc, SignedMessage
from nacl.bindings import crypto_sign, crypto_sign_BYTES
import nacl.encoding


def verify_sign(public_key: bytes, signed_message: bytes, signature: bytes):
    key = VerifyKey(public_key)
    try:
        key.verify(signed_message, signature)
        return True
    except exc.BadSignatureError:
        return False


def sign_message(message: bytes,
                 signing_key,
                 encoder: nacl.encoding.Encoder = nacl.encoding.RawEncoder, ) -> bytes:
    raw_signed = crypto_sign(message, signing_key)

    signature = encoder.encode(raw_signed[:crypto_sign_BYTES])
    message = encoder.encode(raw_signed[crypto_sign_BYTES:])
    signed = encoder.encode(raw_signed)

    return SignedMessage._from_parts(signature, message, signed).signature



"""
WIP
"""
import os
import re
import typing

from bitarray import bitarray
from bitarray.util import hex2ba

from pytoniq.boc.slice import Slice


class TlbGeneratorError(BaseException):
    pass


class TlbDeserializatorError(BaseException):
    pass


class TlbSchema:

    def __init__(self, class_name: str, class_args: list, fields: dict, constructor_name: str, tag: bitarray):
        self._class_name = class_name
        self._class_args = class_args
        self._fields = fields
        self._constructor_name = constructor_name
        self._tag = tag

    @property
    def constructor_name(self) -> str:
        return self._constructor_name

    @property
    def class_name(self) -> str:
        return self._class_name

    @property
    def class_args(self) -> list:
        return self._class_args

    @property
    def tag(self) -> bitarray:
        return self._tag

    @property
    def fields(self) -> typing.Dict[str, str]:
        return self._fields

    def parse_class_args(self):
        pass

    def __repr__(self):
        return f'<TlbSchema#{self._tag.tobytes().hex() if self._tag else "_"} {self._class_name} {"".join(self._class_args)} with fields {self._fields}>'
        # return f'<TlbSchema#{self._tag.tobytes().hex() if self._tag else "_"} {self._class_name} {"".join(self._class_args)}>'


class TlbSchemas:

    def __init__(self, schemas: typing.List[TlbSchema]) -> None:
        self._list: typing.List[TlbSchema] = schemas
        self.class_name_map: typing.Dict[str, typing.List[TlbSchema]] = {}
        self._lexer = Lexer()
        self.generate_map()

    def append(self, sch: TlbSchema) -> None:
        self._list.append(sch)
        self.class_name_map[sch.class_name] = self.class_name_map.get(sch.class_name, []) + [sch]

    def generate_map(self) -> None:
        for schema in self._list:
            self.class_name_map[schema.class_name] = self.class_name_map.get(schema.class_name, []) + [schema]

    def get_by_class_name(self, class_name: str) -> typing.List[TlbSchema]:
        return self.class_name_map.get(class_name)

    @staticmethod
    def _is_int(num: str):
        try:
            int(num)
            return True
        except:
            return False

    @staticmethod
    def _is_eq(param: str) -> bool:
        if '+' in param or '-' in param:
            return True
        return False

    @staticmethod
    def _solve_eq(equation: str, result: int) -> int:
        return None
        # return solve(f'Eq({equation}, {result})')[0]

    @staticmethod
    def _find_var(equation: str) -> typing.Optional[str]:
        for i in equation:
            if i.isalpha():
                return i

    def _find_by_class_args(self, schemas: typing.List[TlbSchema], vars_: list):
        """
        for every possible variant we assign points, and we choose the variant with the max points
        """
        max_point = -1
        best_schema = None
        for sch in schemas:
            cur_point = 0
            if len(sch.class_args) != len(vars_):
                continue
            for i in range(len(sch.class_args)):
                if self._is_int(sch.class_args[i]):
                    if sch.class_args[i] == str(vars_[i]):
                        cur_point += 1
                    else:
                        cur_point -= 1
                elif self._is_eq(sch.class_args[i]):
                    print('eq found: ', sch)
                    # TODO
                    pass
            if cur_point > max_point:
                max_point = cur_point
                best_schema = sch
        if best_schema is None:
            raise TlbDeserializatorError('couldn\'t find fit schema')
        return best_schema

    @staticmethod
    def _compare_by_bit_prefix(cell: Slice, tag: bitarray) -> bool:
        print('comparing', cell.preload_bits(len(tag)), tag)
        return cell.preload_bits(len(tag)) == tag

    def deserialize_field(self, cell_slice: Slice, type_: str, args: dict = {}):
        if type_ == 'Cell':
            return cell_slice
        elif 'bits' in type_:
            return cell_slice.load_bits(int(type_.split('bits')[-1]))
        elif 'int' in type_:
            return cell_slice.load_int(int(type_.split('int')[-1]))
        elif 'uint' in type_:
            return cell_slice.load_uint(int(type_.split('uint')[-1]))
        elif type_.startswith('(##'):                      # (## number)
            return cell_slice.load_uint(int(type_[4:-1]))  # 0123number-1
        elif type_.startswith('(#<='):
            pass  # TODO
        elif type_.startswith('(') and type_.endswith(')'):
            # dict:(HashmapE 32 (VarUInteger 32))
            # stack:(VmStackList depth)
            type_ = type_[1:-1]
            type_ = self._lexer.split(type_)
            type_, class_args = type_[0], type_[1:]
            for i in range(len(class_args)):
                if class_args[i] in args:
                    class_args[i] = args[class_args[i]]
            return self.deserialize(cell_slice, type_, class_args)
        elif type_.startswith('^'):
            return self.deserialize_field(cell_slice.load_ref(), type_[1:], args)
        else:
            return self.deserialize(cell_slice, type_, [])

    def deserialize(self, cell_slice: Slice, class_name: str, class_args: list):
        """
        class_args id dict where keys are indexes or names of arguments and values are arguments values
        for e.g. we have VmStackList
        """
        schemas = self.get_by_class_name(class_name)
        res_schemas = []
        if len(schemas) > 1:  # if we have many possible schemas we compare by tag
            for sch in schemas:
                if self._compare_by_bit_prefix(cell_slice, sch.tag):
                    res_schemas.append(sch)
            schemas = res_schemas
        if len(schemas) > 1:  # if we still have many possible schemas we compare by class arguments
            schemas = [self._find_by_class_args(schemas, class_args)]
        schema = schemas[0]
        if schema.tag:
            cell_slice.load_bits(len(schema.tag))

        result = {}

        for i in range(len(schema.class_args)):
            arg = schema.class_args[i]
            if self._is_eq(arg):
                result[self._find_var(arg)] = self._solve_eq(arg, class_args[i])

        for field, type_ in schema.fields.items():
            result[field] = self.deserialize_field(cell_slice, type_, result)
            continue
            if type_ == 'Cell':
                result[field] = cell_slice
            elif 'bits' in type_:
                result[field] = cell_slice.load_bits(int(type_.split('bits')[-1]))
            elif 'int' in type_:
                result[field] = cell_slice.load_int(int(type_.split('int')[-1]))
            elif 'uint' in type_:
                result[field] = cell_slice.load_uint(int(type_.split('uint')[-1]))
            elif type_.startswith('(##'):                         # (## number)
                result[field] = cell_slice.load_uint(int(type_[4:-1]))  # 0123number-1
            elif type_.startswith('(#<='):
                pass   # TODO
            elif type_.startswith('(') and type_.endswith(')'):
                # dict:(HashmapE 32 (VarUInteger 32))
                # stack:(VmStackList depth)
                type_ = type_[1:-1]
                type_ = self._lexer.split(type_)
                type_, args = type_[0], type_[1:]
                for i in range(len(args)):
                    if args[i] in result:
                        args[i] = result[args[i]]
                result[field] = self.deserialize(cell_slice, type_, args)
            elif type_.startswith('^'):
                result[field] = self.deserialize(cell_slice.load_ref(), type_[1:], class_args)
        return result


class Lexer:

    def __init__(self):
        self._tag_hex = re.compile(r'#([0-9a-f]+_?|_)')
        self._tag_bin = re.compile(r'\$([01]*_?)')
        # https://www.debuggex.com/r/kT4s0-gThkHLZCGO ; to avoid recursion in reg ex and use built-in "re" lib.
        self._splitter = re.compile(r'[\w]*[^\w\s]*\[[^\]]*\]|'
                                    r'[\w]*[^\w\s]*\([^)(]*(?:\([^)(]*(?:\([^)(]*(?:\([^)(]*\)[^)(]*)*\)[^)(]*)*\)[^)(]*)*\)|'
                                    r'[\w]*[^\w\s]*\{[^\}]*\}|'
                                    r'\S+')

    def detect_tag(self, constructor: str) -> typing.Tuple[str, bitarray]:
        """
        :param constructor: constructor name with tag
        :return: constructor name and tag in bitarray
        """
        # hex_tag = self._tag_hex.findall(constructor)
        # if not hex_tag:
        #     bin_tag = self._tag_bin.findall(constructor)
        #     if not bin_tag:
        #         return constructor, bitarray('')
        #     return bitarray(bin_tag[0])
        # return hex2ba(hex_tag[0], 'big')
        if '#' in constructor:
            cons_name, tag = constructor.split('#')
            tag = tag.replace('_', '')
            return cons_name, hex2ba(tag, 'big')
        if '$' in constructor:
            cons_name, tag = constructor.split('$')
            tag = tag.replace('_', '')
            return cons_name, bitarray(tag)
        return constructor, bitarray('')  # or crc32? TODO

    def split(self, s: str):
        return self._splitter.findall(s)


class TlbRegistrator:

    def __init__(self):
        self._lexer = Lexer()

    def register(self, schema: str) -> TlbSchema:
        if '=' not in schema:
            raise TlbGeneratorError('can\'t parse tlb schema: expected "=" in schema')
        splited = self._lexer.split(schema.replace(';', ''))
        eq_i = splited.index('=')

        class_name = splited[eq_i + 1]
        class_args = splited[eq_i + 2:]
        fields = self.parse_args(splited[1: eq_i])
        constructor_name, tag = self._lexer.detect_tag(splited[0])

        return TlbSchema(class_name, class_args, fields, constructor_name, tag)

        print('schema: ', schema)
        print('classname: ', class_name)
        print('class args: ', class_args)
        print('constructor name: ', constructor_name)
        print('constructor tag: ', tag)
        print('args: ', fields, '\n\n')

    @staticmethod
    def parse_args(fields: typing.List[str]) -> dict:
        result = {}
        for field in fields:
            if '{' in field or '}' in field:
                continue
            if field.startswith('^'):
                result['_'] = field
                continue
            if ':' not in field:
                name, type_ = '_', field
            else:
                name, type_ = field.split(':', maxsplit=1)
            result[name] = type_
        return result


class TlbGenerator:
    def __init__(self, path: str, registrator: typing.Optional[TlbRegistrator] = None) -> None:
        self._path = os.path.normpath(path)
        if registrator is None:
            registrator = TlbRegistrator()
        self._registrator = registrator

    def generate(self):
        result = []
        if os.path.isdir(self._path):
            for f in os.listdir(self._path):
                if f.endswith('.tlb'):
                    result += self.from_file(os.path.join(self._path, f))
        else:
            result = self.from_file(self._path)
        return TlbSchemas(result)

    def from_file(self, file_path: str):
        result = []
        with open(file_path, 'r') as f:
            temp = ''
            comment = False

            for line in f:
                stripped = line.lstrip()
                if '*/' in stripped:
                    comment = False
                    continue

                if not stripped or stripped.startswith('//') or comment:
                    continue

                if '/*' in stripped:
                    comment = True
                    continue

                if '//' in stripped:
                    stripped = stripped.split('//')[0].strip()

                stripped = stripped.replace('\n', ' ')

                if ';' not in stripped:
                    temp += stripped
                    continue
                else:
                    stripped = temp + stripped
                    temp = ''
                result.append(self._registrator.register(stripped))
                # result.append(stripped)
        return result


if __name__ == '__main__':
    schemas = TlbGenerator('schemas').generate()

    # schemas.deserialize()

    # for i, j in schemas.class_name_map.items():
    #     print(i, j)



from ..boc.deserialize import NullCell
from ..boc.address import Address
from abc import ABC, abstractmethod


def is_builtin_class_instance(obj):
    return obj.__class__.__module__ == '__builtins__'


class TlbError(BaseException):
    pass


class TlbScheme(ABC):
    """
    abstract class for Tlb Schemes wrappers
    """
    @abstractmethod
    def serialize(self, *args): ...

    @classmethod
    @abstractmethod
    def deserialize(cls, *args): ...

    def __repr__(self, t=1):
        # s = f'< Tl-B {self.__class__.__name__} \n'
        # for k, v in self.__dict__.items():
        #     if isinstance(v, (int, str, float, bytes, bool, dict, tuple, list, NullCell, Address, type(None))):
        #         s += '\t' * t + k + ': ' + v.__repr__() + '\n'
        #     else:
        #         s += '\t' * t + k + ': ' + v.__repr__(t + 1) + '\n'
        # return s + '\t' * t + '>' + '\n'
        return f'< Tl-B {self.__class__.__name__} {" ".join([i + ": " + j.__repr__() for i, j in self.__dict__.items()])} >'
        # TODO beautiful repr



import typing

from .block import CurrencyCollection
from .tlb import TlbScheme, TlbError
from .account import AccountStatus, StateInit, StorageUsedShort
from .utils import HashUpdate
from .. import HashMap
from ..boc import Slice, Builder, Cell
from ..boc.address import Address


class TransactionError(TlbError):
    pass


class Transaction(TlbScheme):
    """
    transaction$0111 account_addr:bits256 lt:uint64
    prev_trans_hash:bits256 prev_trans_lt:uint64 now:uint32
    outmsg_cnt:uint15
    orig_status:AccountStatus end_status:AccountStatus
    ^[
        in_msg:(Maybe ^(Message Any))
        out_msgs:(HashmapE 15 ^(Message Any))
    ]
    total_fees:CurrencyCollection state_update:^(HASH_UPDATE Account)
    description:^TransactionDescr = Transaction;

    """

    def __init__(self,
                 account_addr: bytes,
                 lt: int,
                 prev_trans_hash: bytes,
                 prev_trans_lt: int,
                 now: int,
                 outmsg_cnt: int,
                 orig_status: AccountStatus,
                 end_status: AccountStatus,
                 in_msg: typing.Optional["MessageAny"],
                 out_msgs: typing.List["MessageAny"],
                 total_fees: CurrencyCollection,
                 state_update: HashUpdate,
                 description: "TransactionDescr",
                 cell: typing.Optional["Cell"] = None
                 ):
        self.account_addr = account_addr
        self.account_addr_hex = account_addr.hex()
        self.lt = lt
        self.prev_trans_hash = prev_trans_hash
        self.prev_trans_lt = prev_trans_lt
        self.now = now
        self.outmsg_cnt = outmsg_cnt
        self.orig_status = orig_status
        self.end_status = end_status
        self.in_msg = in_msg
        self.out_msgs = out_msgs
        self.total_fees = total_fees
        self.state_update = state_update
        self.description: TransactionOrdinary = description
        # the type hinting above is wrong, but it was made for better experience in IDE cause most time you work with ordinary trs
        self.cell = cell.copy()

    def serialize(self) -> Cell:
        in_msg_cell = Builder()
        if self.in_msg:
            in_msg_cell.store_maybe_ref(self.in_msg.serialize())
        else:
            in_msg_cell.store_bit_int(0)

        def serializer(src, dest):
            dest.store_ref(src.serialize())

        out_msgs = {i: self.out_msgs[i] for i in range(len(self.out_msgs))}
        out_msg_cell = Builder().store_dict(HashMap(key_size=15, map_=out_msgs, value_serializer=serializer).serialize()).end_cell()

        builder = Builder()
        builder\
            .store_bits('0111')\
            .store_bytes(self.account_addr)\
            .store_uint(self.lt, 64)\
            .store_bytes(self.prev_trans_hash)\
            .store_uint(self.prev_trans_lt)\
            .store_uint(self.now, 32)\
            .store_uint(self.outmsg_cnt, 15)\
            .store_cell(self.orig_status.serialize())\
            .store_cell(self.end_status.serialize())\
            .store_ref(
                Builder()\
                    .store_cell(in_msg_cell.end_cell())\
                    .store_cell(out_msg_cell)\
                    .end_cell()
            )\
            .store_cell(self.total_fees.serialize())\
            .store_ref(self.state_update.serialize())\
            .store_ref(self.description.serialize())
            # TODO description serialization

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        cell = cell_slice.copy().to_cell()
        if cell_slice.is_special():
            return cell_slice.to_cell()
        tag = cell_slice.load_bits(4).to01()
        if tag != '0111':
            raise TransactionError(f'Transaction deserialization error unknown prefix tag: {tag}')
        account_addr = cell_slice.load_bytes(32)
        lt = cell_slice.load_uint(64)
        prev_trans_hash = cell_slice.load_bytes(32)
        prev_trans_lt = cell_slice.load_uint(64)
        now = cell_slice.load_uint(32)
        outmsg_cnt = cell_slice.load_uint(15)
        orig_status = AccountStatus.deserialize(cell_slice)
        end_status = AccountStatus.deserialize(cell_slice)
        ref = cell_slice.load_ref().begin_parse()
        in_msg = None
        if ref.load_bit():
            in_msg = MessageAny.deserialize(ref.load_ref().begin_parse())
        out_msgs = ref.load_dict(15, value_deserializer=lambda src: MessageAny.deserialize(src.load_ref().begin_parse()))
        if out_msgs is not None:
            out_msgs = [out_msgs[i] for i in sorted(out_msgs)]
        else:
            out_msgs = []
        total_fees = CurrencyCollection.deserialize(cell_slice)
        state_update = HashUpdate.deserialize(cell_slice.load_ref().begin_parse())
        description = TransactionDescr.deserialize(cell_slice.load_ref().begin_parse())

        return cls(account_addr, lt, prev_trans_hash, prev_trans_lt, now, outmsg_cnt, orig_status, end_status, in_msg, out_msgs, total_fees, state_update, description, cell=cell)


""" ########## MESSAGES ########## """


class MessageAny(TlbScheme):
    """
    message$_ {X:Type} info:CommonMsgInfo
    init:(Maybe (Either StateInit ^StateInit))
    body:(Either X ^X) = Message X;
    """
    def __init__(self, info: typing.Union["InternalMsgInfo", "ExternalMsgInfo", "ExternalOutMsgInfo"],
                 init: typing.Optional[StateInit], body: Cell):
        self.info = info
        self.init = init
        self.body = body

    @property
    def is_external(self):
        return isinstance(self.info, ExternalMsgInfo)

    @property
    def is_internal(self):
        return isinstance(self.info, InternalMsgInfo)

    @property
    def is_external_out(self):
        return isinstance(self.info, ExternalOutMsgInfo)

    def serialize(self) -> Cell:
        builder = Builder().store_cell(self.info.serialize())
        if self.init:
            builder.store_bit(1)  # maybe true
            builder.store_bit(1)  # Either right
            builder.store_ref(self.init.serialize())
        else:
            builder.store_bit(0)  # maybe false
        builder.store_maybe_ref(self.body)  # Either right
        return builder.end_cell()

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        info = CommonMsgInfo.deserialize(cell_slice)
        init = None
        maybe = cell_slice.load_bit()
        if maybe:
            either = cell_slice.load_bit()
            if either:  # right
                init = StateInit.deserialize(cell_slice.load_ref().begin_parse())
            else:  # left
                init = StateInit.deserialize(cell_slice)
        either = cell_slice.load_bit()
        if either:  # right
            body = cell_slice.load_ref()
        else:  # left
            body = cell_slice.to_cell()
        return cls(info, init, body)


class CommonMsgInfo(TlbScheme):
    """
    int_msg_info$0 ihr_disabled:Bool bounce:Bool bounced:Bool
    src:MsgAddressInt dest:MsgAddressInt
    value:CurrencyCollection ihr_fee:Grams fwd_fee:Grams
    created_lt:uint64 created_at:uint32 = CommonMsgInfo;

    ext_in_msg_info$10 src:MsgAddressExt dest:MsgAddressInt
    import_fee:Grams = CommonMsgInfo;

    ext_out_msg_info$11 src:MsgAddressInt dest:MsgAddressExt
    created_lt:uint64 created_at:uint32 = CommonMsgInfo;
    """
    def __init__(self):
        pass

    @classmethod
    def serialize(cls, *args):
        pass

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        tag = cell_slice.preload_bit()
        if not tag:  # 0
            return InternalMsgInfo.deserialize(cell_slice)
        tag = cell_slice.preload_bits(2).to01()
        if tag == '10':
            return ExternalMsgInfo.deserialize(cell_slice)
        # 11
        return ExternalOutMsgInfo.deserialize(cell_slice)


class InternalMsgInfo(CommonMsgInfo):
    """
    int_msg_info$0 ihr_disabled:Bool bounce:Bool bounced:Bool
    src:MsgAddressInt dest:MsgAddressInt
    value:CurrencyCollection ihr_fee:Grams fwd_fee:Grams
    created_lt:uint64 created_at:uint32 = CommonMsgInfo;
    """
    def __init__(self, ihr_disabled: bool, bounce: bool, bounced: bool, src: Address, dest: Address,
                 value: CurrencyCollection, ihr_fee: int, fwd_fee: int, created_lt: int, created_at: int):
        super().__init__()
        self.ihr_disabled = ihr_disabled
        self.bounce = bounce
        self.bounced = bounced
        self.src = src
        self.dest = dest
        self.value = value
        self.value_coins: int = value.grams
        self.ihr_fee = ihr_fee
        self.fwd_fee = fwd_fee
        self.created_lt = created_lt
        self.created_at = created_at

    def serialize(self) -> Cell:
        builder = Builder()
        builder.store_uint(0, 1)  # $0
        return builder\
            .store_bool(self.ihr_disabled)\
            .store_bool(self.bounce)\
            .store_bool(self.bounced)\
            .store_address(self.src)\
            .store_address(self.dest)\
            .store_cell(self.value.serialize())\
            .store_coins(self.ihr_fee)\
            .store_coins(self.fwd_fee)\
            .store_uint(self.created_lt, 64)\
            .store_uint(self.created_at, 32)\
            .end_cell()

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        tag = cell_slice.load_bit()
        if tag:
            raise TransactionError(f'InternalMsgInfo deserialization error unknown prefix tag: {tag}')
        return cls(
            ihr_disabled=cell_slice.load_bool(),
            bounce=cell_slice.load_bool(),
            bounced=cell_slice.load_bool(),
            src=cell_slice.load_address(),
            dest=cell_slice.load_address(),
            value=CurrencyCollection.deserialize(cell_slice),
            ihr_fee=cell_slice.load_coins(),
            fwd_fee=cell_slice.load_coins(),
            created_lt=cell_slice.load_uint(64),
            created_at=cell_slice.load_uint(32)
        )


class ExternalMsgInfo(CommonMsgInfo):
    """
    ext_in_msg_info$10 src:MsgAddressExt dest:MsgAddressInt
    import_fee:Grams = CommonMsgInfo;
    """
    def __init__(self, src: Address, dest: Address, import_fee: int):
        super().__init__()
        self.src = src
        self.dest = dest
        self.import_fee = import_fee

    def serialize(self) -> Cell:
        builder = Builder()
        builder.store_uint(2, 2)  # $10
        return builder\
            .store_address(self.src)\
            .store_address(self.dest)\
            .store_coins(self.import_fee)\
            .end_cell()

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        tag = cell_slice.load_bits(2).to01()
        if tag != '10':
            raise TransactionError(f'ExternalMsgInfo deserialization error unknown prefix tag: {tag}')
        return cls(
            src=cell_slice.load_address(),
            dest=cell_slice.load_address(),
            import_fee=cell_slice.load_coins()
        )


class ExternalOutMsgInfo(CommonMsgInfo):
    """
    ext_out_msg_info$11 src:MsgAddressInt dest:MsgAddressExt
    created_lt:uint64 created_at:uint32 = CommonMsgInfo;
    """
    def __init__(self, src: Address, dest: Address, created_lt: int, created_at: int):
        super().__init__()
        self.src = src
        self.dest = dest
        self.created_lt = created_lt
        self.created_at = created_at

    def serialize(self) -> Cell:
        builder = Builder()
        builder.store_uint(3, 2)  # $11
        return builder\
            .store_address(self.src)\
            .store_address(self.dest)\
            .store_uint(self.created_lt, 64)\
            .store_uint(self.created_at, 32)\
            .end_cell()

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        tag = cell_slice.load_bits(2).to01()
        if tag != '11':
            raise TransactionError(f'ExternalOutMsgInfo deserialization error unknown prefix tag: {tag}')
        return cls(
            src=cell_slice.load_address(),
            dest=cell_slice.load_address(),
            created_lt=cell_slice.load_uint(64),
            created_at=cell_slice.load_uint(32)
        )


""" ########## PHASES ########## """


class AccStatusChange(TlbScheme):
    """
    acst_unchanged$0 = AccStatusChange;  // x -> x
    acst_frozen$10 = AccStatusChange;    // init -> frozen
    acst_deleted$11 = AccStatusChange;   // frozen -> deleted
    """
    def __init__(self, type_: typing.Literal["unchanged", "frozen", "deleted"]):
        self.type_ = type_

    @classmethod
    def serialize(cls, *args):
        pass

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        tag = cell_slice.load_bit()
        if not tag:
            return cls('unchanged')
        tag = cell_slice.load_bit()
        if not tag:
            return cls('frozen')
        return cls('deleted')


class ComputeSkipReason(TlbScheme):
    """
    cskip_no_state$00 = ComputeSkipReason;
    cskip_bad_state$01 = ComputeSkipReason;
    cskip_no_gas$10 = ComputeSkipReason;
    cskip_suspended$110 = ComputeSkipReason;
    """
    def __init__(self, type_: typing.Literal["no_state", "bad_state", "no_gas", "suspended"]):
        self.type_ = type_

    @classmethod
    def serialize(cls, *args):
        pass

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        tag = cell_slice.load_bit()
        if not tag:
            tag = cell_slice.load_bit()
            if not tag:  # 00
                return cls('no_state')
            # 01
            return cls('bad_state')
        tag = cell_slice.load_bit()
        if not tag:  # 10
            return cls('no_gas')
        tag = cell_slice.load_bit()
        if not tag:  # 110
            return cls('suspended')
        raise TransactionError(f'ComputeSkipReason deserialization error unknown prefix tag: {111}')


class TrStoragePhase(TlbScheme):
    """
    tr_phase_storage$_ storage_fees_collected:Grams
    storage_fees_due:(Maybe Grams)
    status_change:AccStatusChange
    = TrStoragePhase;
    """
    def __init__(self,
                 storage_fees_collected: int,
                 storage_fees_due: typing.Optional[int],
                 status_change: AccStatusChange
                 ):
        self.storage_fees_collected = storage_fees_collected
        self.storage_fees_due = storage_fees_due
        self.status_change = status_change

    @classmethod
    def serialize(cls, *args):
        pass

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        return cls(
            storage_fees_collected=cell_slice.load_coins(),
            storage_fees_due=cell_slice.load_coins() if cell_slice.load_bit() else None,
            status_change=AccStatusChange.deserialize(cell_slice)
        )


class TrCreditPhase(TlbScheme):
    """
    tr_phase_credit$_ due_fees_collected:(Maybe Grams)
    credit:CurrencyCollection = TrCreditPhase;
    """
    def __init__(self,
                 due_fees_collected: typing.Optional[int],
                 credit: CurrencyCollection,
                 ):
        self.due_fees_collected = due_fees_collected
        self.credit = credit

    @classmethod
    def serialize(cls, *args):
        pass

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        return cls(
            due_fees_collected=cell_slice.load_coins() if cell_slice.load_bit() else None,
            credit=CurrencyCollection.deserialize(cell_slice),
        )


class TrComputePhase(TlbScheme):
    """
    tr_phase_compute_skipped$0 reason:ComputeSkipReason = TrComputePhase;

    tr_phase_compute_vm$1 success:Bool msg_state_used:Bool
    account_activated:Bool gas_fees:Grams
    ^[
        gas_used:(VarUInteger 7)
        gas_limit:(VarUInteger 7) gas_credit:(Maybe (VarUInteger 3))
        mode:int8 exit_code:int32 exit_arg:(Maybe int32)
        vm_steps:uint32
        vm_init_state_hash:bits256 vm_final_state_hash:bits256
    ] = TrComputePhase;
    """
    def __init__(self,
                 type_: typing.Literal["skipped", "vm"],
                 reason: typing.Optional[ComputeSkipReason] = None,
                 success: bool = None,
                 msg_state_used: bool = None,
                 account_activated: bool = None,
                 gas_fees: int = None,
                 gas_used: int = None,
                 gas_limit: int = None,
                 gas_credit: typing.Optional[int] = None,
                 mode: int = None,
                 exit_code: int = None,
                 exit_arg: typing.Optional[int] = None,
                 vm_steps: int = None,
                 vm_init_state_hash: bytes = None,
                 vm_final_state_hash: bytes = None
                 ):
        self.type_ = type_
        if type_ == 'skipped':
            self.reason = reason
            return
        self.success = success
        self.msg_state_used = msg_state_used
        self.account_activated = account_activated
        self.gas_fees = gas_fees
        self.gas_used = gas_used
        self.gas_limit = gas_limit
        self.gas_credit = gas_credit
        self.mode = mode
        self.exit_code = exit_code
        self.exit_arg = exit_arg
        self.vm_steps = vm_steps
        self.vm_init_state_hash = vm_init_state_hash
        self.vm_final_state_hash = vm_final_state_hash

    @classmethod
    def serialize(cls, *args):
        pass

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        """
        tr_phase_compute_vm$1 success:Bool msg_state_used:Bool
        account_activated:Bool gas_fees:Grams
        ^[
            gas_used:(VarUInteger 7)
            gas_limit:(VarUInteger 7) gas_credit:(Maybe (VarUInteger 3))
            mode:int8 exit_code:int32 exit_arg:(Maybe int32)
            vm_steps:uint32
            vm_init_state_hash:bits256 vm_final_state_hash:bits256
        ] = TrComputePhase;
        """
        tag = cell_slice.load_bit()
        if not tag:
            return cls('skipped', ComputeSkipReason.deserialize(cell_slice))
        success = cell_slice.load_bool()
        msg_state_used = cell_slice.load_bool()
        account_activated = cell_slice.load_bool()
        gas_fees = cell_slice.load_coins()
        ref = cell_slice.load_ref().begin_parse()
        gas_used = ref.load_var_uint(3)  # int(7).bit_length()
        gas_limit = ref.load_var_uint(3)  # int(7).bit_length()
        gas_credit = ref.load_var_uint(2) if ref.load_bit() else None
        mode = ref.load_int(8)
        exit_code = ref.load_int(32)
        exit_arg = ref.load_int(32) if ref.load_bit() else None
        vm_steps = ref.load_uint(32)
        vm_init_state_hash = ref.load_bytes(32)
        vm_final_state_hash = ref.load_bytes(32)
        return cls(
            type_='vm',
            reason=None,
            success=success,
            msg_state_used=msg_state_used,
            account_activated=account_activated,
            gas_fees=gas_fees,
            gas_used=gas_used,
            gas_limit=gas_limit,
            gas_credit=gas_credit,
            mode=mode,
            exit_code=exit_code,
            exit_arg=exit_arg,
            vm_steps=vm_steps,
            vm_init_state_hash=vm_init_state_hash,
            vm_final_state_hash=vm_final_state_hash,
        )


class TrActionPhase(TlbScheme):
    """
    tr_phase_action$_ success:Bool valid:Bool no_funds:Bool
    status_change:AccStatusChange
    total_fwd_fees:(Maybe Grams) total_action_fees:(Maybe Grams)
    result_code:int32 result_arg:(Maybe int32) tot_actions:uint16
    spec_actions:uint16 skipped_actions:uint16 msgs_created:uint16
    action_list_hash:bits256 tot_msg_size:StorageUsedShort
    = TrActionPhase;
    """
    def __init__(self,
                 success: bool,
                 valid: bool,
                 no_funds: bool,
                 status_change: AccStatusChange,
                 total_fwd_fees: typing.Optional[int],
                 total_action_fees: typing.Optional[int],
                 result_code: int,
                 result_arg: typing.Optional[int],
                 tot_actions: int,
                 spec_actions: int,
                 skipped_actions: int,
                 msgs_created: int,
                 action_list_hash: bytes,
                 tot_msg_size: StorageUsedShort):
        self.success = success
        self.valid = valid
        self.no_funds = no_funds
        self.status_change = status_change
        self.total_fwd_fees = total_fwd_fees
        self.total_action_fees = total_action_fees
        self.result_code = result_code
        self.result_arg = result_arg
        self.tot_actions = tot_actions
        self.spec_actions = spec_actions
        self.skipped_actions = skipped_actions
        self.msgs_created = msgs_created
        self.action_list_hash = action_list_hash
        self.tot_msg_size = tot_msg_size

    @classmethod
    def serialize(cls, *args):
        pass

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        return cls(
            success=cell_slice.load_bool(),
            valid=cell_slice.load_bool(),
            no_funds=cell_slice.load_bool(),
            status_change=AccStatusChange.deserialize(cell_slice),
            total_fwd_fees=cell_slice.load_coins() if cell_slice.load_bit() else None,
            total_action_fees=cell_slice.load_coins() if cell_slice.load_bit() else None,
            result_code=cell_slice.load_int(32),
            result_arg=cell_slice.load_int(32) if cell_slice.load_bit() else None,
            tot_actions=cell_slice.load_uint(16),
            spec_actions=cell_slice.load_uint(16),
            skipped_actions=cell_slice.load_uint(16),
            msgs_created=cell_slice.load_uint(16),
            action_list_hash=cell_slice.load_bytes(32),
            tot_msg_size=StorageUsedShort.deserialize(cell_slice)
        )


class TrBouncePhase(TlbScheme):
    """
    tr_phase_bounce_negfunds$00 = TrBouncePhase;

    tr_phase_bounce_nofunds$01 msg_size:StorageUsedShort
    req_fwd_fees:Grams = TrBouncePhase;

    tr_phase_bounce_ok$1 msg_size:StorageUsedShort
    msg_fees:Grams fwd_fees:Grams = TrBouncePhase;
    """
    def __init__(self,
                 type_: typing.Literal["negfunds", "nofunds", "ok"],
                 msg_size: typing.Optional[StorageUsedShort] = None,
                 req_fwd_fees: typing.Optional[int] = None,
                 msg_fees: typing.Optional[int] = None,
                 fwd_fees: typing.Optional[int] = None
                 ):
        self.type_ = type_
        if type_ == 'negfunds':
            return
        if type_ == 'nofunds':
            self.msg_size = msg_size
            self.req_fwd_fees = req_fwd_fees
            return
        self.msg_size = msg_size
        self.msg_fees = msg_fees
        self.fwd_fees = fwd_fees

    @classmethod
    def serialize(cls, *args):
        pass

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        tag = cell_slice.load_bit()
        if tag:  # 1
            return cls(
                'ok',
                msg_size=StorageUsedShort.deserialize(cell_slice),
                msg_fees=cell_slice.load_coins(),
                fwd_fees=cell_slice.load_coins()
            )
        tag = cell_slice.load_bit()
        if tag:  # 01
            return cls(
                'nofunds',
                msg_size=StorageUsedShort.deserialize(cell_slice),
                req_fwd_fees=cell_slice.load_coins(),
            )
        # 00
        return cls('negfunds')


""" ########## TR DESCRIPTIONS ########## """


class TransactionDescr(TlbScheme):
    """
    trans_ord$0000 credit_first:Bool storage_ph:(Maybe TrStoragePhase) credit_ph:(Maybe TrCreditPhase)
    compute_ph:TrComputePhase action:(Maybe ^TrActionPhase) aborted:Bool bounce:(Maybe TrBouncePhase)
    destroyed:Bool = TransactionDescr;

    trans_storage$0001 storage_ph:TrStoragePhase = TransactionDescr;

    trans_tick_tock$001 is_tock:Bool storage_ph:TrStoragePhase compute_ph:TrComputePhase
     action:(Maybe ^TrActionPhase) aborted:Bool destroyed:Bool = TransactionDescr;

    trans_split_prepare$0100 split_info:SplitMergeInfo storage_ph:(Maybe TrStoragePhase)
    compute_ph:TrComputePhase action:(Maybe ^TrActionPhase) aborted:Bool destroyed:Bool
    = TransactionDescr;

    trans_split_install$0101 split_info:SplitMergeInfo prepare_transaction:^Transaction
    installed:Bool = TransactionDescr;

    trans_merge_prepare$0110 split_info:SplitMergeInfo storage_ph:TrStoragePhase aborted:Bool
    = TransactionDescr;

    trans_merge_install$0111 split_info:SplitMergeInfo prepare_transaction:^Transaction
    storage_ph:(Maybe TrStoragePhase) credit_ph:(Maybe TrCreditPhase)
    compute_ph:TrComputePhase action:(Maybe ^TrActionPhase) aborted:Bool destroyed:Bool
    = TransactionDescr;
    """
    def __init__(self):
        ...

    @classmethod
    def serialize(cls, *args):
        pass

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        tag = cell_slice.load_bits(3).to01()
        if tag == '001':
            return TransactionTickTock.deserialize(cell_slice)

        tag += str(cell_slice.load_bit())
        if tag == '0000':
            return TransactionOrdinary.deserialize(cell_slice)
        if tag == '0001':
            return TransactionStorage.deserialize(cell_slice)
        if tag == '0100':
            return TransactionSplitPrepare.deserialize(cell_slice)
        if tag == '0101':
            return TransactionSplitInstall.deserialize(cell_slice)
        if tag == '0110':
            return TransactionMergePrepare.deserialize(cell_slice)
        if tag == '0111':
            return TransactionMergeInstall.deserialize(cell_slice)

        raise TransactionError(f'TransactionDescr deserialization error unexpected tag: {tag}')


class TransactionOrdinary(TlbScheme):
    """
    trans_ord$0000 credit_first:Bool storage_ph:(Maybe TrStoragePhase) credit_ph:(Maybe TrCreditPhase)
    compute_ph:TrComputePhase action:(Maybe ^TrActionPhase) aborted:Bool bounce:(Maybe TrBouncePhase)
    destroyed:Bool = TransactionDescr;
    """

    def __init__(self,
                 credit_first: bool,
                 storage_ph: typing.Optional["TrStoragePhase"],
                 credit_ph: typing.Optional["TrCreditPhase"],
                 compute_ph: typing.Optional["TrComputePhase"],
                 action: typing.Optional["TrActionPhase"],
                 aborted: bool,
                 bounce: typing.Optional["TrBouncePhase"],
                 destroyed: bool):
        self.type_ = 'ordinary'
        self.credit_first = credit_first
        self.storage_ph = storage_ph
        self.credit_ph = credit_ph
        self.compute_ph = compute_ph
        self.action = action
        self.aborted = aborted
        self.bounce = bounce
        self.destroyed = destroyed

    def serialize(self):
        pass

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        return cls(
            credit_first=cell_slice.load_bool(),
            storage_ph=TrStoragePhase.deserialize(cell_slice) if cell_slice.load_bit() else None,
            credit_ph=TrCreditPhase.deserialize(cell_slice) if cell_slice.load_bit() else None,
            compute_ph=TrComputePhase.deserialize(cell_slice),
            action=TrActionPhase.deserialize(cell_slice.load_ref().begin_parse()) if cell_slice.load_bit() else None,
            aborted=cell_slice.load_bool(),
            bounce=TrBouncePhase.deserialize(cell_slice) if cell_slice.load_bit() else None,
            destroyed=cell_slice.load_bool()
        )


class TransactionStorage(TlbScheme):
    """
    trans_storage$0001 storage_ph:TrStoragePhase = TransactionDescr;
    """
    def __init__(self,
                 storage_ph: TrStoragePhase,
                 ):
        self.type_ = 'storage'
        self.storage_ph = storage_ph

    @classmethod
    def serialize(cls, *args):
        pass

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        return cls(
            storage_ph=TrStoragePhase.deserialize(cell_slice),
        )


class TransactionTickTock(TlbScheme):
    """
    trans_tick_tock$001 is_tock:Bool storage_ph:TrStoragePhase compute_ph:TrComputePhase
    action:(Maybe ^TrActionPhase) aborted:Bool destroyed:Bool = TransactionDescr;
    """
    def __init__(self,
                 is_tock: bool,
                 storage_ph: TrStoragePhase,
                 compute_ph: TrComputePhase,
                 action: typing.Optional[TrActionPhase],
                 aborted: bool,
                 destroyed: bool):
        self.type_ = 'tick_tock'
        self.is_tock = is_tock
        self.storage_ph = storage_ph
        self.compute_ph = compute_ph
        self.action = action
        self.aborted = aborted
        self.destroyed = destroyed

    @classmethod
    def serialize(cls, *args):
        pass

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        return cls(
            is_tock=cell_slice.load_bool(),
            storage_ph=TrStoragePhase.deserialize(cell_slice),
            compute_ph=TrComputePhase.deserialize(cell_slice),
            action=TrActionPhase.deserialize(cell_slice.load_ref().begin_parse()) if cell_slice.load_bit() else None,
            aborted=cell_slice.load_bool(),
            destroyed=cell_slice.load_bool()
        )


class SplitMergeInfo(TlbScheme):
    """
    split_merge_info$_ cur_shard_pfx_len:(## 6)
    acc_split_depth:(## 6) this_addr:bits256 sibling_addr:bits256
    = SplitMergeInfo;
    """
    def __init__(self,
                 cur_shard_pfx_len: int,
                 acc_split_depth: int,
                 this_addr: bytes,
                 sibling_addr: bytes):
        self.cur_shard_pfx_len = cur_shard_pfx_len
        self.acc_split_depth = acc_split_depth
        self.this_addr = this_addr.hex()
        self.sibling_addr = sibling_addr.hex()

    @classmethod
    def serialize(cls, *args):
        ...

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        return cls(
            cur_shard_pfx_len=cell_slice.load_uint(6),
            acc_split_depth=cell_slice.load_uint(6),
            this_addr=cell_slice.load_bytes(32),
            sibling_addr=cell_slice.load_bytes(32)
        )


class TransactionSplitPrepare(TlbScheme):
    """
    trans_split_prepare$0100 split_info:SplitMergeInfo storage_ph:(Maybe TrStoragePhase)
    compute_ph:TrComputePhase action:(Maybe ^TrActionPhase) aborted:Bool destroyed:Bool
    = TransactionDescr;
    """
    def __init__(self,
                 split_info: SplitMergeInfo,
                 storage_ph: typing.Optional[TrStoragePhase],
                 compute_ph: TrComputePhase,
                 action: typing.Optional[TrActionPhase],
                 aborted: bool,
                 destroyed: bool):
        self.type_ = 'split_prepare'
        self.split_info = split_info
        self.storage_ph = storage_ph
        self.compute_ph = compute_ph
        self.action = action
        self.aborted = aborted
        self.destroyed = destroyed

    @classmethod
    def serialize(cls, *args):
        pass

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        return cls(
            split_info=SplitMergeInfo.deserialize(cell_slice),
            storage_ph=TrStoragePhase.deserialize(cell_slice) if cell_slice.load_bit() else None,
            compute_ph=TrComputePhase.deserialize(cell_slice),
            action=TrActionPhase.deserialize(cell_slice.load_ref().begin_parse()) if cell_slice.load_bit() else None,
            aborted=cell_slice.load_bool(),
            destroyed=cell_slice.load_bool()
        )


class TransactionSplitInstall(TlbScheme):
    """
    trans_split_install$0101 split_info:SplitMergeInfo
    prepare_transaction:^Transaction installed:Bool
    = TransactionDescr;
    """
    def __init__(self,
                 split_info: SplitMergeInfo,
                 prepare_transaction: Transaction,
                 installed: bool):
        self.type_ = 'split_install'
        self.split_info = split_info
        self.prepare_transaction = prepare_transaction
        self.installed = installed

    @classmethod
    def serialize(cls, *args):
        pass

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        return cls(
            split_info=SplitMergeInfo.deserialize(cell_slice),
            prepare_transaction=Transaction.deserialize(cell_slice.load_ref().begin_parse()),
            installed=cell_slice.load_bool()
        )


class TransactionMergePrepare(TlbScheme):
    """
    trans_merge_prepare$0110 split_info:SplitMergeInfo
    storage_ph:TrStoragePhase aborted:Bool
    = TransactionDescr;
    """
    def __init__(self,
                 split_info: SplitMergeInfo,
                 storage_ph: TrStoragePhase,
                 aborted: bool
                 ):
        self.type_ = 'merge_prepare'
        self.split_info = split_info
        self.storage_ph = storage_ph
        self.aborted = aborted

    @classmethod
    def serialize(cls, *args):
        pass

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        return cls(
            split_info=SplitMergeInfo.deserialize(cell_slice),
            storage_ph=TrStoragePhase.deserialize(cell_slice),
            aborted=cell_slice.load_bool()
        )


class TransactionMergeInstall(TlbScheme):
    """
    trans_merge_install$0111 split_info:SplitMergeInfo
    prepare_transaction:^Transaction storage_ph:(Maybe TrStoragePhase)
    credit_ph:(Maybe TrCreditPhase)
    compute_ph:TrComputePhase action:(Maybe ^TrActionPhase)
    aborted:Bool destroyed:Bool
    = TransactionDescr;
    """
    def __init__(self,
                 split_info: SplitMergeInfo,
                 prepare_transaction: Transaction,
                 storage_ph: typing.Optional[TrStoragePhase],
                 credit_ph: typing.Optional[TrCreditPhase],
                 compute_ph: TrComputePhase,
                 action: typing.Optional[TrActionPhase],
                 aborted: bool,
                 destroyed: bool
                 ):
        self.type_ = 'merge_install'
        self.split_info = split_info
        self.prepare_transaction = prepare_transaction
        self.storage_ph = storage_ph
        self.credit_ph = credit_ph
        self.compute_ph = compute_ph
        self.action = action
        self.aborted = aborted
        self.destroyed = destroyed

    @classmethod
    def serialize(cls, *args):
        pass

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        """
        trans_merge_install$0111 split_info:SplitMergeInfo
        prepare_transaction:^Transaction storage_ph:(Maybe TrStoragePhase)
        credit_ph:(Maybe TrCreditPhase)
        compute_ph:TrComputePhase action:(Maybe ^TrActionPhase)
        aborted:Bool destroyed:Bool
        = TransactionDescr;
        """
        return cls(
            split_info=SplitMergeInfo.deserialize(cell_slice),
            prepare_transaction=Transaction.deserialize(cell_slice.load_ref().begin_parse()),
            storage_ph=TrStoragePhase.deserialize(cell_slice) if cell_slice.load_bit() else None,
            credit_ph=TrCreditPhase.deserialize(cell_slice) if cell_slice.load_bit() else None,
            compute_ph=TrComputePhase.deserialize(cell_slice),
            action=TrActionPhase.deserialize(cell_slice.load_ref().begin_parse()) if cell_slice.load_bit() else None,
            aborted=cell_slice.load_bool(),
            destroyed=cell_slice.load_bool()
        )


class IntermediateAddress(TlbScheme):  # TODO: maybe move to account.py
    """
    interm_addr_regular$0 use_dest_bits:(#<= 96)
    = IntermediateAddress;

    interm_addr_simple$10 workchain_id:int8 addr_pfx:uint64
    = IntermediateAddress;

    interm_addr_ext$11 workchain_id:int32 addr_pfx:uint64
    = IntermediateAddress;
    """

    def __init__(self,
                 type_: str,
                 use_dest_bits: typing.Optional[int] = None,
                 workchain_id: typing.Optional[int] = None,
                 addr_pfx: typing.Optional[int] = None
                 ):
        self.type_ = type_
        self.use_dest_bits = use_dest_bits
        self.workchain_id = workchain_id
        self.addr_pfx = addr_pfx

    @classmethod
    def serialize(cls, *args):
        ...

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        tag = cell_slice.load_bit()
        if not tag:  # 0
            return cls('interm_addr_regular', use_dest_bits=cell_slice.load_uint(7))
        tag = cell_slice.load_bit()
        if tag:  # 10
            return cls('interm_addr_simple', workchain_id=cell_slice.load_int(8), addr_pfx=cell_slice.load_uint(64))
        # 11
        return cls('interm_addr_ext', workchain_id=cell_slice.load_int(32), addr_pfx=cell_slice.load_uint(64))


class MsgEnvelope(TlbScheme):
    """
    msg_envelope#4 cur_addr:IntermediateAddress
    next_addr:IntermediateAddress fwd_fee_remaining:Grams
    msg:^(Message Any) = MsgEnvelope;
    """

    def __init__(self,
                 cur_addr: IntermediateAddress,
                 next_addr: IntermediateAddress,
                 fwd_fee_remaining: int,
                 msg: MessageAny):
        self.cur_addr = cur_addr
        self.next_addr = next_addr
        self.fwd_fee_remaining = fwd_fee_remaining
        self.msg = msg

    @classmethod
    def serialize(cls, *args):
        ...

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        tag = cell_slice.load_uint(4)
        if tag != 4:
            raise TransactionError(f'MsgEnvelope deserialization error tag: {tag}')
        return cls(
            cur_addr=IntermediateAddress.deserialize(cell_slice),
            next_addr=IntermediateAddress.deserialize(cell_slice),
            fwd_fee_remaining=cell_slice.load_coins(),
            msg=MessageAny.deserialize(cell_slice.load_ref().begin_parse())
        )


class InMsg(TlbScheme):
    """
    msg_import_ext$000 msg:^(Message Any) transaction:^Transaction = InMsg;

    msg_import_ihr$010 msg:^(Message Any) transaction:^Transaction
        ihr_fee:Grams proof_created:^Cell = InMsg;

    msg_import_imm$011 in_msg:^MsgEnvelope
        transaction:^Transaction fwd_fee:Grams = InMsg;

    msg_import_fin$100 in_msg:^MsgEnvelope
        transaction:^Transaction fwd_fee:Grams = InMsg;

    msg_import_tr$101  in_msg:^MsgEnvelope out_msg:^MsgEnvelope
        transit_fee:Grams = InMsg;

    msg_discard_fin$110 in_msg:^MsgEnvelope transaction_id:uint64
        fwd_fee:Grams = InMsg;

    msg_discard_tr$111 in_msg:^MsgEnvelope transaction_id:uint64
        fwd_fee:Grams proof_delivered:^Cell = InMsg;
    """
    def __init__(self,
                 type_: str,
                 msg: typing.Optional[MessageAny] = None,
                 in_msg: typing.Optional[MsgEnvelope] = None,
                 transaction: typing.Optional[Transaction] = None,
                 **kwargs
                 ):
        self.type_ = type_
        self.msg = msg
        self.in_msg = in_msg
        self.transaction = transaction
        for k, v in kwargs.items():
            setattr(self, k, v)

    @classmethod
    def serialize(cls, *args):
        pass

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        tag = cell_slice.load_bits(3).to01()
        if tag == '000':
            return cls('msg_import_ext',
                       msg=MessageAny.deserialize(cell_slice.load_ref().begin_parse()),
                       transaction=Transaction.deserialize(cell_slice.load_ref().begin_parse())
                       )
        if tag == '010':
            return cls('msg_import_ihr',
                       msg=MessageAny.deserialize(cell_slice.load_ref().begin_parse()),
                       transaction=Transaction.deserialize(cell_slice.load_ref().begin_parse()),
                       ihr_fee=cell_slice.load_coins(),
                       proof_created=cell_slice.load_ref()
                       )
        if tag == '011':
            return cls('msg_import_imm',
                       in_msg=MsgEnvelope.deserialize(cell_slice.load_ref().begin_parse()),
                       transaction=Transaction.deserialize(cell_slice.load_ref().begin_parse()),
                       fwd_fee=cell_slice.load_coins(),
                       )
        if tag == '100':
            return cls('msg_import_fin',
                       in_msg=MsgEnvelope.deserialize(cell_slice.load_ref().begin_parse()),
                       transaction=Transaction.deserialize(cell_slice.load_ref().begin_parse()),
                       fwd_fee=cell_slice.load_coins(),
                       )
        if tag == '101':
            return cls('msg_import_tr',
                       in_msg=MsgEnvelope.deserialize(cell_slice.load_ref().begin_parse()),
                       out_msg=MsgEnvelope.deserialize(cell_slice.load_ref().begin_parse()),
                       transit_fee=cell_slice.load_coins(),
                       )
        if tag == '110':
            return cls('msg_discard_fin',
                       in_msg=MsgEnvelope.deserialize(cell_slice.load_ref().begin_parse()),
                       transaction_id=cell_slice.load_uint(64),
                       transit_fee=cell_slice.load_coins(),
                       )
        if tag == '111':
            return cls('msg_discard_tr',
                       in_msg=MsgEnvelope.deserialize(cell_slice.load_ref().begin_parse()),
                       transaction_id=cell_slice.load_uint(64),
                       fwd_fee=cell_slice.load_coins(),
                       proof_delivered=cell_slice.load_ref()
                       )
        raise TransactionError(f'InMsg deserialization error: unknown prefix tag {tag}')


class OutMsg(TlbScheme):
    """
    msg_export_ext$000 msg:^(Message Any)
    transaction:^Transaction = OutMsg;

    msg_export_imm$010 out_msg:^MsgEnvelope
    transaction:^Transaction reimport:^InMsg = OutMsg;

    msg_export_new$001 out_msg:^MsgEnvelope
    transaction:^Transaction = OutMsg;

    msg_export_tr$011  out_msg:^MsgEnvelope
    imported:^InMsg = OutMsg;

    msg_export_deq$1100 out_msg:^MsgEnvelope
    import_block_lt:uint63 = OutMsg;

    msg_export_deq_short$1101 msg_env_hash:bits256
    next_workchain:int32 next_addr_pfx:uint64
    import_block_lt:uint64 = OutMsg;

    msg_export_tr_req$111 out_msg:^MsgEnvelope
    imported:^InMsg = OutMsg;

    msg_export_deq_imm$100 out_msg:^MsgEnvelope
    reimport:^InMsg = OutMsg;
    """
    def __init__(self,
                 type_: str,
                 msg: typing.Optional[MessageAny] = None,
                 out_msg: typing.Optional[MsgEnvelope] = None,
                 transaction: typing.Optional[Transaction] = None,
                 **kwargs
                 ):
        self.type_ = type_
        self.msg = msg
        self.out_msg = out_msg
        self.transaction = transaction
        for k, v in kwargs.items():
            setattr(self, k, v)

    @classmethod
    def serialize(cls, *args):
        pass

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        tag = cell_slice.load_bits(3).to01()
        if tag == '000':
            return cls('msg_export_ext',
                       msg=MessageAny.deserialize(cell_slice.load_ref().begin_parse()),
                       transaction=Transaction.deserialize(cell_slice.load_ref().begin_parse())
                       )
        if tag == '010':
            return cls('msg_export_imm',
                       out_msg=MsgEnvelope.deserialize(cell_slice.load_ref().begin_parse()),
                       transaction=Transaction.deserialize(cell_slice.load_ref().begin_parse()),
                       reimport=InMsg.deserialize(cell_slice.load_ref().begin_parse()),
                       )
        if tag == '001':
            return cls('msg_export_new',
                       out_msg=MsgEnvelope.deserialize(cell_slice.load_ref().begin_parse()),
                       transaction=Transaction.deserialize(cell_slice.load_ref().begin_parse()),
                       )
        if tag == '011':
            return cls('msg_export_tr',
                       out_msg=MsgEnvelope.deserialize(cell_slice.load_ref().begin_parse()),
                       imported=InMsg.deserialize(cell_slice.load_ref().begin_parse()),
                       )
        if tag == '100':
            return cls('msg_export_deq_imm',
                       out_msg=MsgEnvelope.deserialize(cell_slice.load_ref().begin_parse()),
                       reimport=InMsg.deserialize(cell_slice.load_ref().begin_parse()),
                       )
        if tag == '111':
            return cls('msg_export_tr_req',
                       out_msg=MsgEnvelope.deserialize(cell_slice.load_ref().begin_parse()),
                       imported=InMsg.deserialize(cell_slice.load_ref().begin_parse()),
                       )
        tag += str(cell_slice.load_bit())
        if tag == '1100':
            return cls('msg_export_deq',
                       out_msg=MsgEnvelope.deserialize(cell_slice.load_ref().begin_parse()),
                       import_block_lt=cell_slice.load_uint(63)
                       )
        if tag == '1101':
            return cls('msg_export_deq',
                       msg_env_hash=cell_slice.load_bytes(32),
                       next_workchain=cell_slice.load_int(32),
                       next_addr_pfx=cell_slice.load_uint(64),
                       import_block_lt=cell_slice.load_uint(64)
                       )
        raise TransactionError(f'OutMsg deserialization error: unknown prefix tag {tag}')


class ImportFees(TlbScheme):
    """
    import_fees$_ fees_collected:Grams
    value_imported:CurrencyCollection = ImportFees;
    """

    def __init__(self, fees_collected: int, value_imported: CurrencyCollection):
        self.fees_collected = fees_collected
        self.value_imported = value_imported

    @classmethod
    def serialize(cls, *args):
        ...

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        return cls(fees_collected=cell_slice.load_coins(), value_imported=CurrencyCollection.deserialize(cell_slice))



import base64
import binascii
import copy
import typing

from bitarray.util import ba2int

from .tvm_bitarray import bitarray, TvmBitarray
from .utils import bytes_to_uint
from ..crypto.crc import crc32c


class BocError(BaseException):
    pass


# https://github.com/ton-blockchain/ton/blob/24dc184a2ea67f9c47042b4104bbb4d82289fac1/crypto/tl/boc.tlb#L25
SERIALIZED_BOC_IDX_CRC32C = b'\xac\xc3\xa7('  # LEAN_BOC_MAGIC_PREFIX_CRC acc3a728
SERIALIZED_BOC_IDX_PREFIX = b'h\xffe\xf3'  # LEAN_BOC_MAGIC_PREFIX 68ff65f3
SERIALIZED_BOC_PREFIX = b'\xb5\xee\x9cr'  # REACH_BOC_MAGIC_PREFIX b5ee9c72


class NullCell:
    """
    NullCell is a class that stores cell bits, and refs to other NullCells.
    This class needed to deserialize boc (bag of cells) and convert it into Cell, Builder or Slice
    """
    def __init__(self, bits: TvmBitarray, refs: list, type_: int):
        self.bits = bits
        self.refs = refs
        self.type_ = type_

    def get_refs(self, cls: type):
        refs = []
        for ref in self.refs:
            refs.append(cls(ref.bits, ref.get_refs(cls), ref.type_))

        return refs

    def to_cell(self):
        from .cell import Cell
        return Cell(self.bits, self.get_refs(Cell), self.type_)

    def to_slice(self):
        from .slice import Slice
        from .cell import Cell
        return Slice(self.bits, self.get_refs(Cell), )

    def to_builder(self):
        from .builder import Builder
        cell = self.to_cell()
        return Builder().store_cell(cell)

    def copy(self):
        return NullCell(self.bits.copy(), copy.deepcopy(self.refs.copy()), self.type_)

    def __repr__(self) -> str:
        return f'<NullCell {len(self.bits)}[{self.bits.tobytes().hex().upper()}] -> {len(self.refs)} refs>'

    def __str__(self, t=1, comma=False) -> str:
        """
        :param t: \t symbols amount before text
        :param comma: "," after "}"
        """
        text = ('* ' * (self.type_ != -1)) + f'{len(self.bits)}[{self.bits.tobytes().hex().upper()}]'
        if self.refs:
            text += f' -> {{\n'
            for index, ref in enumerate(self.refs):
                next_comma = True if index != len(self.refs) - 1 else False
                text += '\t' * t + ref.__str__(t + 1, next_comma) + '\n'
            text += '\t' * (t - 1) + '}'
        if comma:
            text += ','
        return text


class Boc:

    def __init__(self, data: typing.Union[bytes, str]):
        if not isinstance(data, bytes):
            try:
                data = bytes.fromhex(data)
            except ValueError:
                try:
                    data = base64.b64decode(data)
                except binascii.Error:
                    raise BocError('boc data in unknown form')
        self.data = data
        self.data_len = len(data)

    @classmethod
    def from_base64(cls, data: str):
        return cls(base64.b64decode(data))

    @classmethod
    def from_hex(cls, data: str):
        return cls(bytes.fromhex(data))

    def parse(self):
        pass

    @staticmethod
    def deserialize_boc_header(data: bytes):
        # TODO: check if this implementation is really faster than bytes slicing
        data_len = len(data)
        if data_len < 4:
            raise BocError(f'not enough bytes to deserialize boc header: {data}')
        result = {
            'has_idx': True,
            'hash_crc32': None,
            'has_cache_bits': False,
            'flags': 0,
            'size_bytes': data[0],
            'offset_bytes': None,
            'cells_num': None,
            'roots_num': None,
            'absent_num': None,
            'tot_cells_size': None,
            'root_list': None,
            'index': None,
            'cells_data': None,
        }
        if data[:4] == SERIALIZED_BOC_PREFIX:
            flags_byte = data[4]
            result['has_idx'] = flags_byte & 128
            result['hash_crc32'] = flags_byte & 64
            result['has_cache_bits'] = flags_byte & 32
            result['flags'] = (flags_byte & 16) * 2 + (flags_byte & 8)
            result['size_bytes'] = flags_byte % 8
        elif data[:4] == SERIALIZED_BOC_IDX_PREFIX:
            result['hash_crc32'] = 0
        elif data[:4] == SERIALIZED_BOC_IDX_CRC32C:
            result['hash_crc32'] = 1
        else:
            raise BocError(f'unknown boc prefix: {data[:4]}')
        if data_len - 5 < 1 + 5 * result['size_bytes']:
            raise BocError(f'can\'t parse boc header: {data[:4]}')
        offset_bytes = data[5]
        result['offset_bytes'] = offset_bytes
        size_bytes = result['size_bytes']

        end = 6 + 3 * size_bytes
        result['cells_num'], result['roots_num'], result['absent_num'] \
            = [bytes_to_uint(data[i: i + size_bytes]) for i in range(6, end, size_bytes)]

        i = end + result['offset_bytes']
        result['tot_cells_size'] = bytes_to_uint(data[end: i])

        if data_len - i < result['roots_num'] * size_bytes:
            raise Exception("Not enough bytes for encoding root cells hashes")
        end = i + result['roots_num'] * size_bytes
        result['root_list'] = [bytes_to_uint(data[j: j + size_bytes]) for j in range(i, end,  size_bytes)]
        i = end
        if result['has_idx']:
            if data_len - i < offset_bytes * result['cells_num']:
                raise BocError("Not enough bytes for index encoding")
            end = i + result['cells_num'] * offset_bytes
            result['index'] = [bytes_to_uint(data[j: j + offset_bytes]) for j in range(i, end,  offset_bytes)]
            i = end

        if data_len - i < result['tot_cells_size']:
            raise BocError("Not enough bytes for cells data")

        end = i + result['tot_cells_size']
        result['cells_data'] = data[i: end]
        i = end

        if result['hash_crc32']:
            if data_len - i < 4:
                raise BocError("Not enough bytes for crc32c hashsum")
            if crc32c(data[: i]) != data[i: i + 4]:
                raise BocError("Crc32c hashsum mismatch")
            i += 4
        if data_len - i:  # != 0
            raise BocError("Too many bytes in boc")
        return result

    @staticmethod
    def deserialize_cell(data: bytes, ref_index_size: int) -> typing.Tuple[dict, int]:
        data_len = len(data)
        refs_descriptor = data[0]
        level = refs_descriptor >> 5
        total_refs = refs_descriptor & 7
        has_hashes = (refs_descriptor & 16) != 0
        is_exotic = refs_descriptor & 8
        is_absent = total_refs == 7 and has_hashes
        if is_absent:
            raise BocError('can\'t deserialize absent cell')
        bits_descriptor = data[1]
        is_augmented = bits_descriptor & 1
        data_size = (bits_descriptor >> 1) + is_augmented
        hashes_size = (level + 1) * 32 if has_hashes else 0
        depth_size = (level + 1) * 2 if hashes_size else 0
        i = 2

        if data_len - i < hashes_size + depth_size + data_size + ref_index_size * total_refs:
            raise BocError('Not enough bytes to encode cell data')

        if has_hashes:
            i += hashes_size + depth_size
        bits = bitarray()
        bits.frombytes(data[i: i + data_size])
        i += data_size

        end = None
        if is_augmented and bits:
            for j in range(-1, -8, -1):
                if bits[j] == 1:
                    end = j
                    break
        bits = TvmBitarray(1023, bits[:end])

        if is_exotic:
            if len(bits) < 8:
                raise BocError('not enough bytes for an exotic cell type')
            cell_type = ba2int(bits[:8], signed=True)
        else:
            cell_type = -1

        cell_refs_indexes = []
        for r in range(total_refs):
            cell_refs_indexes.append(bytes_to_uint(data[i: i + ref_index_size]))
            i += ref_index_size

        # cell = NullCell(bits, cell_refs_indexes, cell_type)
        cell = {'bits': bits, 'refs': cell_refs_indexes, 'type': cell_type, 'result': None}

        return cell, i

    def deserialize(self, cls: type = None):
        if not cls:
            from .cell import Cell
            cls = Cell

        header = self.deserialize_boc_header(self.data)
        cells_data = header['cells_data']
        cells_array = []

        i = 0

        for ci in range(header['cells_num']):
            cell, j = self.deserialize_cell(cells_data[i:], header['size_bytes'])
            i += j
            cells_array.append(cell)

        for ci in reversed(range(header['cells_num'])):
            c = cells_array[ci]
            refs = []
            for ri in range(len(c['refs'])):
                r = c['refs'][ri]
                if r < ci:
                    raise Exception('Topological order is broken')
                refs.append(cells_array[r]['result'])
            cells_array[ci]['result'] = cls(cells_array[ci]['bits'], refs, cells_array[ci]['type'])

        root_cells = []
        for ri in header['root_list']:
            root_cells.append(cells_array[ri]['result'])

        return root_cells



import copy

from .tlb import TlbScheme, TlbError
from ..boc.slice import Slice
from ..boc.builder import Builder
from ..boc.cell import Cell


class VmError(TlbError):
    pass


class VmStack(TlbScheme):
    """
    vm_stack#_ depth:(## 24) stack:(VmStackList depth) = VmStack;
    """
    @classmethod
    def serialize(cls, data: list) -> "Cell":
        result = Builder()
        result.store_uint(len(data), 24)  # depth
        return result.store_cell(VmStackList.serialize(data.copy())).end_cell()

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        depth = cell_slice.load_uint(24)
        return VmStackList.deserialize(cell_slice, depth)


class VmStackList(TlbScheme):
    """
    vm_stk_cons#_ {n:#} rest:^(VmStackList n) tos:VmStackValue = VmStackList (n + 1);
    vm_stk_nil#_ = VmStackList 0;
    """
    @classmethod
    def serialize(cls, data: list) -> Cell:
        builder = Builder()
        if len(data) == 0:
            return builder.end_cell()
        value = data.pop()
        builder.store_ref(cls.serialize(data))
        return builder.store_cell(VmStackValue.serialize(value)).end_cell()

    @classmethod
    def deserialize(cls, cell_slice: Slice, n_p_1: int):  # n_p_1 stands for n plus 1 or n + 1
        if n_p_1 == 0:
            return []
        result = cls.deserialize(cell_slice.load_ref().begin_parse(), n_p_1 - 1)
        return result + [VmStackValue.deserialize(cell_slice)]


class VmStackValue(TlbScheme):
    """
    vm_stk_null#00 = VmStackValue;
    vm_stk_tinyint#01 value:int64 = VmStackValue;
    vm_stk_int#0201_ value:int257 = VmStackValue;
    vm_stk_nan#02ff = VmStackValue;
    vm_stk_cell#03 cell:^Cell = VmStackValue;
    vm_stk_slice#04 _:VmCellSlice = VmStackValue;
    vm_stk_builder#05 cell:^Cell = VmStackValue;
    vm_stk_cont#06 cont:VmCont = VmStackValue;
    vm_stk_tuple#07 len:(## 16) data:(VmTuple len) = VmStackValue;
    """

    @classmethod
    def serialize(cls, value) -> Cell:
        builder = Builder()
        if value is None:
            builder.store_bytes(b'\x00')
        elif isinstance(value, int):
            if value.bit_length() <= 64:
                builder.store_bytes(b'\x01')
                builder.store_int(value, 64)
            else:
                builder.store_bits('000000100000000')  # 0201_
                # builder.store_bytes(b'\x02\x01')
                builder.store_int(value, 257)
        elif isinstance(value, bytes):
            builder.store_bits('0000001000000000')  # 0200
            assert len(value) <= 32, 'bytes length should be less than 32'
            builder.store_bytes(value)
        elif isinstance(value, Cell):
            builder.store_bytes(b'\x03')
            builder.store_ref(value)
        elif isinstance(value, Slice):
            builder.store_bytes(b'\x04')
            builder.store_cell(VmCellSlice.serialize(value))
        elif isinstance(value, Builder):
            builder.store_bytes(b'\x05')
            builder.store_ref(value.end_cell())
        elif isinstance(value, VmCont):
            builder.store_bytes(b'\x06')
            builder.store_cell(VmCont.serialize(value))
        elif isinstance(value, VmTuple):
            builder.store_bytes(b'\x07')
            builder.store_uint(len(value), 16)
            builder.store_cell(VmTuple.serialize(value))
        return builder.end_cell()

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        tag = cell_slice.preload_bits(15).to01()
        if tag == '000000100000000':  # #0201_
            cell_slice.load_bits(15)
            return cell_slice.load_int(257)
            # actually the code below is more convenient (for keys for e.g.) but may work incorrectly for a big positive integer
            if cell_slice.preload_bit():
                return cell_slice.load_int(257)
            cell_slice.load_bit()
            return cell_slice.load_bytes(32)

        tag = cell_slice.preload_bytes(2)
        if tag[:1] == b'\x00':
            cell_slice.load_bytes(1)
            return None
        elif tag[:1] == b'\x01':
            cell_slice.load_bytes(1)
            return cell_slice.load_int(64)
        elif tag == b'\x02\xff':
            cell_slice.load_bytes(2)
            return None
        elif tag[:1] == b'\x03':
            cell_slice.load_bytes(1)
            return cell_slice.load_ref()
        elif tag[:1] == b'\x05':
            cell_slice.load_bytes(1)
            return cell_slice.load_ref().to_builder()
        elif tag[:1] == b'\x04':
            cell_slice.load_bytes(1)
            return VmCellSlice.deserialize(cell_slice)
        elif tag[:1] == b'\x06':
            cell_slice.load_bytes(1)
            return VmCont.deserialize(cell_slice)
        elif tag[:1] == b'\x07':
            cell_slice.load_bytes(1)
            tuple_len = cell_slice.load_uint(16)
            return VmTuple.deserialize(cell_slice, tuple_len)


class VmTuple(TlbScheme):
    """
    vm_tuple_nil$_ = VmTuple 0;
    vm_tuple_tcons$_ {n:#} head:(VmTupleRef n) tail:^VmStackValue = VmTuple (n + 1);
    """
    """
    Usage:
        tuple = VmTuple([1, 2, <Slice 32[00000203] -> 2 refs>])
        VmTuple.serialize(tuple) -> Cell
        tuple = VmTuple.deserialize(cell_slice: Slice, length: int) -> <VmTuple [1, 2, <Slice 32[00000203] -> 2 refs>]>
        tuple.list -> [1, 2, <Slice 32[00000203] -> 2 refs>]
    """

    def __init__(self, list_: list):
        self.list: list = list_

    def __add__(self, other: "VmTuple"):
        self.list += other.list

    def __call__(self, index: int):
        return self.list[index]

    def __getitem__(self, index: int):
        return self.__call__(index)

    def __len__(self):
        return len(self.list)

    def __repr__(self):
        return f'<VmTuple {self.list} >'

    def append(self, item):
        self.list.append(item)
        return self

    def pop(self, index: int = -1):
        return self.list.pop(index)

    @classmethod
    def serialize(cls, values: "VmTuple") -> Cell:
        if len(values) == 0:
            return Cell.empty()
        builder = Builder()
        value = values.pop()
        builder.store_cell(VmTupleRef.serialize(values))
        builder.store_ref(VmStackValue.serialize(value))
        return builder.end_cell()

    @classmethod
    def deserialize(cls, cell_slice: Slice, length: int) -> "VmTuple":
        if length == 0:
            return VmTuple([])

        return VmTupleRef.deserialize(cell_slice, length - 1).append(VmStackValue.deserialize(cell_slice.load_ref().begin_parse()))


class VmTupleRef(TlbScheme):
    """
    vm_tupref_nil$_ = VmTupleRef 0;
    vm_tupref_single$_ entry:^VmStackValue = VmTupleRef 1;
    vm_tupref_any$_ {n:#} ref:^(VmTuple (n + 2)) = VmTupleRef (n + 2);
    """

    @classmethod
    def serialize(cls, values: VmTuple) -> Cell:
        if len(values) == 0:
            return Cell.empty()
        if len(values) == 1:
            return Builder().store_ref(VmStackValue.serialize(values[0])).end_cell()
        return Builder().store_ref(VmTuple.serialize(values)).end_cell()

    @classmethod
    def deserialize(cls, cell_slice: Slice, length: int) -> VmTuple:
        if length == 0:
            return VmTuple([])
        if length == 1:
            return VmTuple([VmStackValue.deserialize(cell_slice.load_ref().begin_parse())])
        return VmTuple.deserialize(cell_slice.load_ref().begin_parse(), length)


class VmCellSlice(TlbScheme):
    """
    _ cell:^Cell st_bits:(## 10) end_bits:(## 10) { st_bits <= end_bits }
    st_ref:(#<= 4) end_ref:(#<= 4) { st_ref <= end_ref } = VmCellSlice;
    """

    @classmethod
    def serialize(cls, value: Slice) -> Cell:
        builder = Builder()
        builder.store_ref(Builder().store_slice(value).end_cell())
        builder.store_uint(0, 10)  # st_bits
        builder.store_uint(len(value.bits), 10)  # end_bits
        builder.store_uint(0, 3)  # st_ref
        builder.store_uint(len(value.refs) - value.ref_offset, 3)  # end_ref
        return builder.end_cell()

    @classmethod
    def deserialize(cls, cell_slice: Slice) -> Slice:
        cell = cell_slice.load_ref()
        st_bits = cell_slice.load_uint(10)
        end_bits = cell_slice.load_uint(10)
        if not st_bits <= end_bits:
            raise VmError(f'{cls.__name__} deserialization error: st_bits {st_bits}, end_bits {end_bits}')

        st_ref = cell_slice.load_uint(3)
        end_ref = cell_slice.load_uint(3)

        if not st_ref <= end_ref:
            raise VmError(f'{cls.__name__} deserialization error: st_ref {st_ref}, end_ref {end_ref}')

        result = Slice.from_cell(cell)
        result.refs = result.refs[st_ref: end_ref]
        result.bits = result.bits[st_bits: end_bits]

        return result


class VmCont(TlbScheme):
    """
    vmc_std$00 cdata:VmControlData code:VmCellSlice = VmCont;
    vmc_envelope$01 cdata:VmControlData next:^VmCont = VmCont;
    vmc_quit$1000 exit_code:int32 = VmCont;
    vmc_quit_exc$1001 = VmCont;
    vmc_repeat$10100 count:uint63 body:^VmCont after:^VmCont = VmCont;
    vmc_until$110000 body:^VmCont after:^VmCont = VmCont;
    vmc_again$110001 body:^VmCont = VmCont;
    vmc_while_cond$110010 cond:^VmCont body:^VmCont after:^VmCont = VmCont;
    vmc_while_body$110011 cond:^VmCont body:^VmCont after:^VmCont = VmCont;
    vmc_pushint$1111 value:int32 next:^VmCont = VmCont;
    """

    def __init__(self, type_, **kwargs):
        self.type_ = type_
        for k, v in kwargs.items():
            setattr(self, k, v)

    @classmethod
    def serialize(cls, value: "VmCont") -> "Cell":
        builder = Builder()

        if value.type_ == 'vmc_std':
            builder.store_bits('00')
            builder.store_cell(VmControlData.serialize(value.cdata))  # cdata:VmControlData
            builder.store_cell(VmCellSlice.serialize(value.code))  # code:VmCellSlice
        elif value.type_ == 'vmc_envelope':
            builder.store_bits('01')
            builder.store_cell(VmControlData.serialize(value.cdata))  # cdata:VmControlData
            builder.store_ref(cls.serialize(value.next))  # next:^VmCont
        elif value.type_ == 'vmc_quit':
            builder.store_bits('1000')
            builder.store_int(value.exit_code, 32)  # exit_code:int32
        elif value.type_ == 'vmc_quit_exc':
            builder.store_bits('1001')
        elif value.type_ == 'vmc_repeat':
            builder.store_bits('10100')
            builder.store_uint(value.count, 63)  # count:uint63
            builder.store_ref(cls.serialize(value.body))  # body:^VmCont
            builder.store_ref(cls.serialize(value.after))  # after:^VmCont
        elif value.type_ == 'vmc_until':
            builder.store_bits('110000')
            builder.store_ref(cls.serialize(value.body))  # body:^VmCont
            builder.store_ref(cls.serialize(value.after))  # after:^VmCont
        elif value.type_ == 'vmc_again':
            builder.store_bits('110001')
            builder.store_ref(cls.serialize(value.body))  # body:^VmCont
        elif value.type_ == 'vmc_while_cond':
            builder.store_bits('110010')
            builder.store_ref(cls.serialize(value.cond))  # cond:^VmCont
            builder.store_ref(cls.serialize(value.body))  # body:^VmCont
            builder.store_ref(cls.serialize(value.after))  # after:^VmCont
        elif value.type_ == 'vmc_while_body':
            builder.store_bits('110011')
            builder.store_ref(cls.serialize(value.cond))  # cond:^VmCont
            builder.store_ref(cls.serialize(value.body))  # body:^VmCont
            builder.store_ref(cls.serialize(value.after))  # after:^VmCont
        elif value.type_ == 'vmc_pushint':
            builder.store_bits('1111')
            builder.store_int(value.value, 32)  # value:int32
            builder.store_ref(cls.serialize(value.next))  # next:^VmCont
        return builder.end_cell()

    @classmethod
    def deserialize(cls, cell_slice: Slice) -> "VmCont":
        tag = cell_slice.preload_bits(6).to01()
        if tag[:2] == '00':
            return cls('vmc_std', cdata=VmControlData.deserialize(cell_slice), code=VmCellSlice.deserialize(cell_slice))
        elif tag[:2] == '01':
            return cls('vmc_envelope', cdata=VmControlData.deserialize(cell_slice), next=cls.deserialize(cell_slice.load_ref().begin_parse()))
        elif tag[:4] == '1000':
            return cls('vmc_quit', exit_code=cell_slice.load_int(32))
        elif tag[:4] == '1001':
            return cls('vmc_quit_exc')
        elif tag[:5] == '10100':
            return cls('vmc_repeat', count=cell_slice.load_uint(63), body=cls.deserialize(cell_slice.load_ref().begin_parse()), after=cls.deserialize(cell_slice.load_ref().begin_parse()))
        elif tag[:6] == '110000':
            return cls('vmc_until', body=cls.deserialize(cell_slice.load_ref().begin_parse()), after=cls.deserialize(cell_slice.load_ref().begin_parse()))
        elif tag[:6] == '110001':
            return cls('vmc_again', body=cls.deserialize(cell_slice.load_ref().begin_parse()))
        elif tag[:6] == '110010':
            return cls('vmc_while_cond', cond=cls.deserialize(cell_slice.load_ref().begin_parse()), body=cls.deserialize(cell_slice.load_ref().begin_parse()), after=cls.deserialize(cell_slice.load_ref().begin_parse()))
        elif tag[:6] == '110011':
            return cls('vmc_while_body', cond=cls.deserialize(cell_slice.load_ref().begin_parse()), body=cls.deserialize(cell_slice.load_ref().begin_parse()), after=cls.deserialize(cell_slice.load_ref().begin_parse()))
        elif tag[:4] == '1111':
            return cls('vmc_pushint', value=cell_slice.load_int(32), next=cls.deserialize(cell_slice.load_ref().begin_parse()))


class VmControlData(TlbScheme):
    """
    vm_ctl_data$_ nargs:(Maybe uint13) stack:(Maybe VmStack) save:VmSaveList
    cp:(Maybe int16) = VmControlData;
    """
    def __init__(self, type_, **kwargs):
        self.type_ = type_
        for k, v in kwargs.items():
            setattr(self, k, v)

    @classmethod
    def serialize(cls, value: "VmControlData") -> Cell:
        builder = Builder()

        if value.nargs:
            builder.store_bit_int(1)
            builder.store_uint(value.nargs, 13)
        else:
            builder.store_bit_int(0)

        if value.stack:
            builder.store_bit_int(1)
            builder.store_cell(value.stack)
        else:
            builder.store_bit_int(0)

        builder.store_cell(VmSaveList.serialize(value.save))

        if value.cp:
            builder.store_bit_int(1)
            builder.store_int(value.cp, 16)
        else:
            builder.store_bit_int(0)

        return builder.end_cell()

    @classmethod
    def deserialize(cls, cell_slice: Slice) -> "VmControlData":
        kwargs = {}
        is_nargs = cell_slice.load_bit()
        if is_nargs:
            kwargs['nargs'] = cell_slice.load_uint(13)
        is_stack = cell_slice.load_bit()
        if is_stack:
            kwargs['stack'] = VmStack.deserialize(cell_slice)
        kwargs['save'] = VmSaveList.deserialize(cell_slice)
        is_cp = cell_slice.load_bit()
        if is_cp:
            kwargs['cp'] = cell_slice.load_int(16)
        return cls('vm_ctl_data', **kwargs)


class VmSaveList(TlbScheme):
    """
    _ cregs:(HashmapE 4 VmStackValue) = VmSaveList;
    """
    @classmethod
    def serialize(cls, value: "HashMap") -> Cell:
        return Builder().store_dict(value).end_cell()

    @classmethod
    def deserialize(cls, cell_slice: Slice) -> "HashMap":
        return cell_slice.load_dict(4)



"""
This is quite an advanced level, so it is highly recommended that you first become familiar with https://docs.ton.org/tvm.pdf 3.1.2 - 3.1.7
"""


class CellTypes:
    ordinary = -1
    pruned_branch = 1
    library_ref = 2
    merkle_proof = 3
    merkle_update = 4


class LevelMask:
    # https://github.com/xssnick/tonutils-go/blob/master/tvm/cell/level.go#L17
    def __init__(self, m: int):
        self._m = m
        self._level = self.get_level()
        self._hash_index = self.get_hash_index()

    @property
    def mask(self):
        return self._m

    @property
    def level(self):
        return self._level

    @property
    def hash_index(self):
        return self._hash_index

    def get_level(self):
        return self._m.bit_length()

    def get_hash_index(self):
        return self._m.bit_count()

    def apply(self, level: int):
        return LevelMask(self._m & ((1 << level) - 1))

    def is_significant(self, level: int):
        return level == 0 or (self._m >> (level - 1)) % 2 != 0




import typing

from .. import Slice, CellTypes
from ..deserialize import NullCell

from bitarray import bitarray


def read_arbitrary_uint(n: int, ser: bitarray) -> typing.Tuple[int, bitarray]:
    x = 0
    for i in range(n):
        x <<= 1
        r = ser.pop(0)
        x += r
    return x, ser


def deserialize_unary(ser: Slice) -> int:
    n = 0
    r = ser.load_bit()
    while r:
        n += 1
        r = ser.load_bit()
    return n


def deserialize_hml(ser: Slice, m: int) -> typing.Tuple[int, bitarray]:
    _type = ''
    k = ser.load_bit()
    if k:
        k = ser.load_bit()
        if k:  # 11
            _type = 'same'
        else:  # 10
            _type = 'long'
    else:  # 0
        _type = 'short'
    if _type == 'short':
        n = deserialize_unary(ser)
        s = ser.load_bits(n)
    elif _type == 'long':
        l = m.bit_length()
        n = ser.load_uint(l)
        s = ser.load_bits(n)
    else:  # same
        v = ser.load_bit()
        l = m.bit_length()
        n = ser.load_uint(l)
        s = bitarray(str(v) * n)
    return n, s


def deserialize_hashmap_node(cs: Slice, m: int, ret_dict: dict, prefix: bitarray) -> None:
    if cs.type_ != CellTypes.ordinary:
        return None
    if m == 0:  # leaf
        if prefix:
            ret_dict[prefix.to01()] = cs
    else:  # fork
        l_prefix, r_prefix = prefix.copy(), prefix.copy()
        l_prefix.append(False)
        r_prefix.append(True)
        parse(
            cs.load_ref().begin_parse(),
            m - 1,
            ret_dict,
            l_prefix,
            )
        parse(
            cs.load_ref().begin_parse(),
            m - 1,
            ret_dict,
            r_prefix,
            )


def deserialize_hashmap_aug_node(cs: Slice, m: int, ret_dict: dict, extras: list, prefix: bitarray, x_deserializer: typing.Callable, y_deserializer: typing.Callable) -> None:
    if m == 0:  # ahmn_leaf
        extras.append(y_deserializer(cs))
        ret_dict[prefix.to01()] = x_deserializer(cs)
    else:  # ahmn_fork
        l_prefix, r_prefix = prefix.copy(), prefix.copy()
        l_prefix.append(False)
        r_prefix.append(True)
        parse_aug(
            cs.load_ref().begin_parse(),
            m - 1,
            ret_dict,
            extras,
            l_prefix,
            x_deserializer,
            y_deserializer
            )
        parse_aug(
            cs.load_ref().begin_parse(),
            m - 1,
            ret_dict,
            extras,
            r_prefix,
            x_deserializer,
            y_deserializer
            )
        extras.append(y_deserializer(cs))


def parse(slice: Slice, key_length: int, ret_dict: dict, prefix: bitarray) -> None:
    l, suffix = deserialize_hml(slice, key_length)
    prefix.extend(suffix)
    m = key_length - l
    deserialize_hashmap_node(slice, m, ret_dict, prefix.copy())


def parse_aug(slice: Slice, key_length: int, ret_dict: dict, extras: list, prefix: bitarray, x_deserializer: typing.Callable, y_deserializer: typing.Callable) -> None:
    if slice.type_ != CellTypes.ordinary:
        return None
    l, suffix = deserialize_hml(slice, key_length)
    prefix.extend(suffix)
    m = key_length - l
    deserialize_hashmap_aug_node(slice, m, ret_dict, extras, prefix.copy(), x_deserializer, y_deserializer)


def parse_hashmap(dict_cell: Slice, key_len: int) -> typing.Optional[dict]:
    result = {}
    parse(dict_cell, key_len, result, bitarray(''))
    return result


def parse_hashmap_aug(dict_cell: Slice, key_len: int, x_deserializer: typing.Callable, y_deserializer: typing.Callable) -> typing.Optional[typing.Tuple[dict, list]]:
    if dict_cell.type_ != CellTypes.ordinary:
        return None
    result = {}
    extras = []
    parse_aug(dict_cell, key_len, result, extras, bitarray(''), x_deserializer, y_deserializer)
    result = {int(i, 2): j for i, j in result.items()}
    return result, extras



from .block import BlockError, Block, BlockInfo, BlockExtra, ShardState, McStateExtra, ShardAccount, ShardStateUnsplit, CurrencyCollection, ValueFlow, ShardAccounts, ShardIdent, ShardDescr
from .account import AccountError, Account, AccountState, SimpleAccount, SimpleAccountState, StorageInfo, AccountStorage, StateInit
from .config import ConfigError, ConfigParam
from .tlb import TlbError, TlbScheme
from .transaction import TransactionError, Transaction, TransactionDescr, TransactionOrdinary, TransactionStorage, TrStoragePhase, TrActionPhase, TrComputePhase, TrBouncePhase, TrCreditPhase, TransactionTickTock, InMsg, OutMsg, InternalMsgInfo, ExternalMsgInfo, ExternalOutMsgInfo, MessageAny
from .vm_stack import VmError, VmStack, VmStackList, VmStackValue, VmSaveList, VmCont, VmTuple, VmTupleRef, VmCellSlice, VmControlData
from .utils import MerkleUpdate, HashUpdate, deserialize_shard_hashes

from .custom import *



import typing

from .tlb import TlbScheme, TlbError
from .utils import MerkleUpdate, HashUpdate
from ..boc import Slice, Builder, Cell
from ..boc.address import Address


class AccountError(TlbError):
    pass


class SimpleAccount(TlbScheme):
    """
    // simple_balance$_ Grams = AccountBalance;

    account_none$0 = SimpleAccount;
    account$1 addr:Address balance:Grams state:SimpleAccountState = SimpleAccount;
    """
    """
    this schema is needed for user-friendly and familiar account representation,
    but not used anywhere outside of this library
    """
    def __init__(self, address: Address, balance: int, state: "SimpleAccountState"):
        self.address = address
        self.balance = balance
        self.state = state

    @classmethod
    def serialize(cls, *args):
        ...

    @classmethod
    def deserialize(cls, *args):
        ...

    @classmethod
    def from_raw(cls, account: "Account", address: typing.Optional[Address] = None):
        if account is None:
            return cls(address=address, balance=0, state=SimpleAccountState.from_raw(None))
        return cls(address=account.addr, balance=account.storage.balance.grams, state=SimpleAccountState.from_raw(account.storage.state))

    def is_uninitialized(self):
        return self.state.type_ == 'uninitialized'

    def is_frozen(self):
        return self.state.type_ == 'frozen'

    def is_active(self):
        return self.state.type_ == 'active'

    def __repr__(self):
        return f'<SimpleAccount {self.address}: state={self.state.type_}, balance={self.balance}>'


class SimpleAccountState(TlbScheme):
    """
    uninitialized$00 = SimpleAccountState;
    frozen$01 state_hash:bits256 = SimpleAccountState;
    active$10 state_init:StateInit = SimpleAccountState;
    """
    """
    this schema is needed for user-friendly and familiar account state representation,
    but not used anywhere outside of this library
    """

    def __init__(self, type_: typing.Literal["uninitialized", "frozen", "active"],
                 state_hash: typing.Optional[bytes] = None,
                 state_init: typing.Optional["StateInit"] = None,
                 ):
        self.type_ = type_
        self.state_hash = state_hash
        self.state_init = state_init

    @classmethod
    def serialize(cls, *args):
        ...

    @classmethod
    def deserialize(cls, *args):
        ...

    @classmethod
    def from_raw(cls, account_state: typing.Optional["AccountState"]):
        if account_state is None or account_state.type_ == 'account_uninit':
            return cls('uninitialized')
        if account_state.type_ == 'account_frozen':
            return cls('frozen', state_hash=account_state.state_hash)
        return cls('active', state_init=account_state.state_init)


class Account(TlbScheme):
    """
    account_none$0 = Account;
    account$1 addr:MsgAddressInt storage_stat:StorageInfo storage:AccountStorage = Account;
    """

    def __init__(self, addr: Address, storage_stat: "StorageInfo", storage: "AccountStorage"):
        self.addr = addr
        self.storage_stat = storage_stat
        self.storage = storage

    @classmethod
    def serialize(cls, *args):
        pass

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        if cell_slice.load_bit():
            return cls(cell_slice.load_address(), StorageInfo.deserialize(cell_slice), AccountStorage.deserialize(cell_slice))
        else:
            return None


class StorageInfo(TlbScheme):
    """
    storage_info$_ used:StorageUsed last_paid:uint32
              due_payment:(Maybe Grams) = StorageInfo;
    """

    def __init__(self, used: "StorageUsed", last_paid: int, due_payment: typing.Optional[int]):
        self.used = used
        self.last_paid = last_paid
        self.due_payment = due_payment

    @classmethod
    def serialize(cls, *args):
        pass

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        return cls(StorageUsed.deserialize(cell_slice), cell_slice.load_uint(32), cell_slice.load_coins() if cell_slice.load_bit() else None)


class StorageUsed(TlbScheme):
    """
    storage_used$_ cells:(VarUInteger 7) bits:(VarUInteger 7)
    public_cells:(VarUInteger 7) = StorageUsed;
    """

    def __init__(self, cells, bits, public_cells):
        self.cells = cells
        self.bits = bits
        self.public_cells = public_cells

    @classmethod
    def serialize(cls, *args):
        pass

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        l = 3  # int(7).bit_length()
        return cls(cell_slice.load_var_uint(l), cell_slice.load_var_uint(l), cell_slice.load_var_uint(l),)


class StorageUsedShort(TlbScheme):
    """
    storage_used_short$_ cells:(VarUInteger 7) bits:(VarUInteger 7) = StorageUsedShort;
    """

    def __init__(self, cells, bits):
        self.cells = cells
        self.bits = bits

    @classmethod
    def serialize(cls, *args):
        pass

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        l = 3  # int(7).bit_length()
        return cls(cell_slice.load_var_uint(l), cell_slice.load_var_uint(l))


class AccountStorage(TlbScheme):
    """
    account_storage$_ last_trans_lt:uint64
    balance:CurrencyCollection state:AccountState
    = AccountStorage;
    """

    def __init__(self, last_trans_lt: int, balance: "CurrencyCollection", state: "AccountState"):
        self.last_trans_lt = last_trans_lt
        self.balance = balance
        self.state = state

    @classmethod
    def serialize(cls, *args):
        pass

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        from .block import CurrencyCollection
        return cls(
            cell_slice.load_uint(64),
            CurrencyCollection.deserialize(cell_slice),
            AccountState.deserialize(cell_slice)
        )


class AccountState(TlbScheme):
    """
    account_uninit$00 = AccountState;
    account_active$1 _:StateInit = AccountState;
    account_frozen$01 state_hash:bits256 = AccountState;
    """

    def __init__(self, type_: str, **kwargs):
        self.type_ = type_
        self.state_init: StateInit
        self.state_hash: bytes
        for k, v in kwargs.items():
            setattr(self, k, v)

    @classmethod
    def serialize(cls, *args):
        pass

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        b = cell_slice.load_bit()
        kwargs = {}
        if b:  # $1
            type_ = 'account_active'
            kwargs['state_init'] = StateInit.deserialize(cell_slice)
        else:
            b = cell_slice.load_bit()
            if b:  # $01
                type_ = 'account_frozen'
                kwargs['state_hash'] = cell_slice.load_bytes(32).hex()
            else:  # $00
                type_ = 'account_uninit'
        return cls(type_, **kwargs)


class StateInit(TlbScheme):
    """
    _ split_depth:(Maybe (## 5)) special:(Maybe TickTock)
    code:(Maybe ^Cell) data:(Maybe ^Cell)
    library:(Maybe ^Cell) = StateInit;
    """

    def __init__(self,
                 split_depth: typing.Optional[int] = None,
                 special: typing.Optional["TickTock"] = None,
                 code: typing.Optional[Cell] = None,
                 data: typing.Optional[Cell] = None,
                 library: typing.Optional[Cell] = None):
        self.split_depth = split_depth
        self.special = special
        self.code = code
        self.data = data
        self.library = library

    def serialize(self) -> Cell:
        builder = Builder()
        builder.store_bit(1).store_uint(self.split_depth, 5) if self.split_depth is not None else builder.store_bit(0)
        builder.store_bit(1).store_cell(self.special.serialize()) if self.special is not None else builder.store_bit(0)
        builder.store_maybe_ref(self.code)
        builder.store_maybe_ref(self.data)
        builder.store_maybe_ref(self.library)
        return builder.end_cell()

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        return cls(
            split_depth=cell_slice.load_uint(5) if cell_slice.load_bit() else None,
            special=TickTock.deserialize(cell_slice) if cell_slice.load_bit() else None,
            code=cell_slice.load_ref() if cell_slice.load_bit() else None,
            data=cell_slice.load_ref() if cell_slice.load_bit() else None,
            library=cell_slice.load_ref() if cell_slice.load_bit() else None,
        )


class TickTock(TlbScheme):
    """
    tick_tock$_ tick:Bool tock:Bool = TickTock;
    """
    def __init__(self, tick: bool, tock: bool):
        self.tick = tick
        self.tock = tock

    def serialize(self) -> Cell:
        return Builder().store_bool(self.tick).store_bool(self.tock).end_cell()

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        return cls(cell_slice.load_bool(), cell_slice.load_bool())


class ShardAccount(TlbScheme):
    """
    account_descr$_ account:^Account last_trans_hash:bits256
    last_trans_lt:uint64 = ShardAccount;
    """
    def __init__(self, account: Account, last_trans_hash: bytes, last_trans_lt: int, cell: typing.Optional[Cell] = None):
        self.account = account
        self.last_trans_hash = last_trans_hash
        self.last_trans_lt = last_trans_lt
        self.cell = cell  # to check merkle proof of account states

    @classmethod
    def serialize(cls, *args):
        pass

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        cell_copy = cell_slice.copy()  # TODO optimize
        return cls(account=Account.deserialize(cell_slice.load_ref().begin_parse()),
                   last_trans_hash=cell_slice.load_bytes(32),
                   last_trans_lt=cell_slice.load_uint(64), cell=cell_copy.to_cell())


class AccountStatus(TlbScheme):
    """
    acc_state_uninit$00 = AccountStatus;
    acc_state_frozen$01 = AccountStatus;
    acc_state_active$10 = AccountStatus;
    acc_state_nonexist$11 = AccountStatus;
    """
    def __init__(self, type_: typing.Literal["uninitialized", "frozen", "active", "nonexist"]):
        self.type_ = type_

    def serialize(self) -> Cell:
        if self.type_ == 'uninitialized':
            return Builder().store_uint(0, 2).end_cell()
        if self.type_ == 'frozen':
            return Builder().store_uint(1, 2).end_cell()
        if self.type_ == 'active':
            return Builder().store_uint(2, 2).end_cell()
        if self.type_ == 'nonexist':
            return Builder().store_uint(3, 2).end_cell()

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        tag = cell_slice.load_bit()
        if tag:
            tag2 = cell_slice.load_bit()
            if tag2:  # 11
                return cls(type_='nonexist')
            else:  # 10
                return cls(type_='active')
        else:
            tag2 = cell_slice.load_bit()
            if tag2:  # 01
                return cls(type_='frozen')
            else:  # 00
                return cls(type_='uninitialized')


class AccountBlock(TlbScheme):
    """
    acc_trans#5 account_addr:bits256
    transactions:(HashmapAug 64 ^Transaction CurrencyCollection)
    state_update:^(HASH_UPDATE Account)
    = AccountBlock;
    """
    def __init__(self,
                 account_addr: str,
                 transactions: typing.Tuple[dict, list],
                 state_update: HashUpdate):
        self.account_addr = account_addr
        self.transactions = transactions
        self.state_update = state_update

    @classmethod
    def serialize(cls, *args):
        pass

    @classmethod
    def deserialize(cls, cell_slice: Slice):
        from .transaction import Transaction, CurrencyCollection
        tag = cell_slice.load_uint(4)
        if tag != 5:
            raise AccountError(f'AccountBlock deserialization error: unknown prefix tag {tag}')
        return cls(
            account_addr=cell_slice.load_bytes(32).hex(),
            transactions=cell_slice.load_hashmap_aug(64, x_deserializer=lambda src: Transaction.deserialize(src.load_ref().begin_parse()), y_deserializer=CurrencyCollection.deserialize),
            state_update=HashUpdate.deserialize(cell_slice.load_ref().begin_parse())
        )



from .slice import Slice
from .cell import Cell, CellError
from .builder import Builder
from .exotic import CellTypes
from .dict import *
from .address import Address, AddressError


def begin_cell():
    return Builder()



import typing
from copy import deepcopy
from bitarray.util import ba2int

from .deserialize import Boc, NullCell
from .cell import Cell
from .tvm_bitarray import bitarray, TvmBitarray, BitarrayLike
from .address import Address


class Slice(NullCell):

    def __init__(self, bits: TvmBitarray, refs: typing.List[Cell], type_: int = -1):
        self.bits = bits.copy()
        self.refs = refs
        self.type_ = type_
        # super().__init__(bits, refs, type_)
        self.ref_offset = 0

    @property
    def remaining_bits(self):
        return len(self.bits)

    @property
    def remaining_refs(self):
        return len(self.refs) - self.ref_offset

    def is_special(self):
        from . import CellTypes
        return False if self.type_ == CellTypes.ordinary else True

    def preload_bit(self) -> int:
        return self.bits[0]

    def load_bit(self) -> int:
        bit = self.preload_bit()
        del self.bits[0]
        return bit

    def preload_bool(self) -> bool:
        return bool(self.bits[0])

    def load_bool(self) -> bool:
        bit = self.preload_bit()
        del self.bits[0]
        return bool(bit)

    def skip_bits(self, length: int) -> "Slice":
        del self.bits[:length]
        return self

    def preload_bits(self, length: int) -> BitarrayLike:
        bits = self.bits[:length]
        return bits

    def load_bits(self, length: int) -> BitarrayLike:
        bits = self.preload_bits(length)
        del self.bits[:length]
        return bits

    def preload_uint(self, length: int) -> int:
        return ba2int(self.bits[:length], signed=False)

    def load_uint(self, length: int) -> int:
        uint = self.preload_uint(length)
        del self.bits[:length]
        return uint

    def preload_int(self, length: int) -> int:
        return ba2int(self.bits[:length], signed=True)

    def load_int(self, length: int) -> int:
        integer = self.preload_int(length)
        del self.bits[:length]
        return integer

    def preload_bytes(self, length: int) -> bytes:
        return self.bits[:length * 8].tobytes()

    def load_bytes(self, length: int) -> bytes:
        bytes_ = self.preload_bytes(length)
        del self.bits[:length * 8]
        return bytes_

    def preload_address(self) -> typing.Optional[Address]:
        # address := flags 2bits, anycast 1bit, workchain 8bits, hash_part 256bits = 267 bits
        if self.preload_uint(2) == 0:
            return None
        rem = self.preload_bits(265)  # 267 - 2

        wc = ba2int(rem[1:9], signed=True)
        hash_part = rem[9:].tobytes()

        return Address((wc, hash_part))

    def load_address(self) -> typing.Optional[Address]:
        # address := flags 2bits, anycast 1bit, workchain 8bits, hash_part 256bits = 267 bits
        if self.load_uint(2) == 0:
            return None
        self.skip_bits(1)
        wc = self.load_int(8)
        hash_part = self.load_bytes(32)
        return Address((wc, hash_part))

    def preload_var_uint(self, bit_length: int) -> int:
        length = self.preload_uint(bit_length)
        if not length:
            return 0
        coins = self.preload_bits(bit_length + length * 8)[bit_length:]
        return ba2int(coins, signed=False)

    def load_var_uint(self, bit_length: int) -> typing.Optional[int]:
        length = self.load_uint(bit_length)
        if not length:
            return None
        return self.load_uint(length * 8)

    def preload_coins(self) -> int:
        length = self.preload_uint(4)
        if not length:
            return 0
        coins = self.preload_bits(4 + length * 8)[4:]
        return ba2int(coins, signed=False)

    def load_coins(self) -> typing.Optional[int]:
        length = self.load_uint(4)
        if not length:
            return 0
        return self.load_uint(length * 8)

    def preload_string(self, byte_length: int = 0):
        if byte_length == 0:
            byte_length = len(self.bits) // 8
        return self.preload_bytes(byte_length)

    def load_string(self, byte_length: int = 0):
        if byte_length == 0:
            byte_length = len(self.bits) // 8
        return self.load_bytes(byte_length)

    def load_snake_bytes(self) -> bytes:
        assert not self.remaining_bits % 8, f'invalid string length: {self.remaining_bits}'
        assert self.remaining_refs in (0, 1), f'invalid amount of refs: {self.remaining_refs}'
        if not self.remaining_refs:
            return self.load_bytes(self.remaining_bits // 8)
        return self.load_bytes(self.remaining_bits // 8) + self.load_ref().begin_parse().load_snake_bytes()

    def load_snake_string(self) -> str:
        return self.load_snake_bytes().decode()

    def preload_ref(self) -> Cell:
        return self.refs[self.ref_offset]

    def load_ref(self) -> Cell:
        ref = self.refs[self.ref_offset]
        self.ref_offset += 1
        return ref

    def preload_maybe_ref(self) -> typing.Optional[Cell]:
        if self.preload_bool():
            return self.refs[self.ref_offset]
        else:
            return None

    def load_maybe_ref(self) -> typing.Optional[Cell]:
        if self.load_bit():
            ref = self.refs[self.ref_offset]
            self.ref_offset += 1
            return ref
        else:
            return None

    def load_hashmap(self, key_length: int, key_deserializer: typing.Callable = None,
                     value_deserializer: typing.Callable = None):
        from .dict.dict import HashMap
        return HashMap.parse(self, key_length, key_deserializer, value_deserializer)

    def load_hashmap_aug(self, key_length: int, x_deserializer: typing.Callable = None,
                         y_deserializer: typing.Callable = None):
        from .dict.parse import parse_hashmap_aug
        return parse_hashmap_aug(self, key_length, x_deserializer, y_deserializer)

    def load_hashmap_aug_e(self, key_length: int, x_deserializer: typing.Callable = None,
                           y_deserializer: typing.Callable = None):
        if self.is_special():
            return self.to_cell()
        if self.load_bit():
            from .dict.parse import parse_hashmap_aug
            return parse_hashmap_aug(self.load_ref().begin_parse(), key_length, x_deserializer, y_deserializer)
        else:
            return {}, [self]  # extra

    def preload_dict(self, key_length: int, key_deserializer: typing.Callable = None,
                     value_deserializer: typing.Callable = None):
        from .dict.dict import HashMap
        if self.preload_bit():
            return HashMap.parse(self.preload_ref().begin_parse(), key_length, key_deserializer, value_deserializer)
        else:
            return None

    def load_dict(self, key_length: int, key_deserializer: typing.Callable = None,
                  value_deserializer: typing.Callable = None):
        from .dict.dict import HashMap
        if self.load_bit():
            return HashMap.parse(self.load_ref().begin_parse(), key_length, key_deserializer, value_deserializer)
        else:
            return None

    @classmethod
    def from_cell(cls, cell: "Cell"):
        return cls(cell.bits, cell.refs, cell.type_)

    @classmethod
    def one_from_boc(cls, data: typing.Any) -> "Slice":
        boc = Boc(data)
        cells = boc.deserialize()
        return cells[0].begin_parse()

    def copy(self):
        return Slice(self.bits.copy(), deepcopy(self.refs), self.type_)

    def __repr__(self) -> str:
        return f'<Slice {len(self.bits)}[{self.bits.tobytes().hex().upper()}] -> {len(self.refs) - self.ref_offset} refs>'

    def __str__(self, t=1, comma=False) -> str:
        """
        :param t: \t symbols amount before text
        :param comma: "," after "}"
        """
        text = f'{len(self.bits)}[{self.bits.tobytes().hex().upper()}]'
        if self.refs:
            text += f' -> {{\n'
            for index, ref in enumerate(self.refs[self.ref_offset:]):
                next_comma = True if index != len(self.refs) - 1 else False
                text += '\t' * t + ref.__str__(t + 1, next_comma) + '\n'
            text += '\t' * (t - 1) + '}'
        if comma:
            text += ','
        return text
